import copy
import itertools
import operator
import sqlite3

# All functions in this file written by Dana Wyman for TALON, and 
# adapted to interface with TALON dbs

# Converts input to string that can be used for IN database query
def format_for_in(l):	
	if type(l) is tuple:
		l = list(l)
	if type(l) is str:
		l = [l]
	return "(" + ','.join(['"' + str(x) + '"' for x in l]) + ")" 

def fetch_all_transcript_gene_pairs(cursor):
	""" Return gene_ID - transcript_ID tuples from database """

	query = """ SELECT gene_ID, transcript_ID FROM transcripts """
	cursor.execute(query)
	
	pairs = cursor.fetchall()
	return pairs
	
def fetch_all_datasets(cursor):
	""" Return a list of all datasets in database """
	cursor.execute("SELECT dataset_name FROM dataset")
	datasets = [str(x[0]) for x in cursor.fetchall()]
	return datasets

def parse_whitelist(whitelist_file):
	""" From the whitelist file, obtain a list of acccepted gene and 
		transcript IDs tuples"""
	whitelist = set()
	with open(whitelist_file, 'r') as f:
		for line in f:
			line = line.strip()
			fields = line.split(",")
			gene_ID = fields[0]
			transcript_ID = fields[1]
			try:
				whitelist.add((int(gene_ID), int(transcript_ID)))
			except:
				raise ValueError("Gene/Transcript IDs in whitelist must be integer TALON IDs")
	return whitelist

def parse_datasets(dataset, cursor):
	""" From the dataset file, obtain a list of acccepted dataset names"""
	db_datasets = fetch_all_datasets(cursor)
	if dataset not in db_datasets:
		raise ValueError("Dataset name '%s' not found in database" % dataset)
	return dataset

def handle_filtering(database, annot, observed, whitelist_file, dataset):
	""" Determines which transcripts to allow in the analysis. This can be done
		in two different ways. If no whitelist is included, then all of the
		transcripts in the database are included (modified by 'observed'
		option). If a whitelist is provided, then transcripts on that list
		will be included (modified by 'observed' option). This can be
		tuned further by providing a dataset file, but this is optional. """

	conn = sqlite3.connect(database)
	conn.row_factory = sqlite3.Row
	cursor = conn.cursor()

	# Get list of datasets to use in run
	if dataset != None:
		datasets = parse_datasets(dataset, cursor)
	elif observed == True:
		datasets = fetch_all_datasets(cursor)
	else:
		datasets = None

	# Get initial transcript whitelist
	if whitelist_file != None:
		whitelist = parse_whitelist(whitelist_file)
	else:
		whitelist = fetch_all_transcript_gene_pairs(cursor)

	if datasets != None:
		# Limit the whitelist to transcripts detected in the datasets
		transcripts = [ x[1] for x in whitelist ]
		transcript_str = format_for_in(transcripts)
		dataset_str = format_for_in(datasets)

		query = """ SELECT DISTINCT gene_ID, transcript_ID
					FROM observed
					WHERE transcript_ID IN %s
					AND dataset in %s """
		cursor.execute(query % (transcript_str, dataset_str))
		whitelist = cursor.fetchall()

	conn.close()
	return whitelist

def get_gene_transcript_map(db, whitelist):
	""" Creates a dictionary mapping gene IDs to the transcripts that belong to
		them. The columns in each tuple are:
			0: gene ID
			1: transcript ID
			2: chromosome
			3: start position (min of 5' and 3')
			4: end position (max of 5' and 3')
			5: strand
			6: edge path
			7. n_exons
	"""

	conn = sqlite3.connect(db)
	conn.row_factory = sqlite3.Row
	cursor = conn.cursor()
	whitelist_string = "(" + ','.join([str(x) for x in whitelist]) + ")"
	query = """
			SELECT 
				t.gene_ID,
				t.transcript_ID,
				loc1.chromosome,
				MIN(loc1.position,loc2.position) AS min_pos,
				MAX(loc1.position,loc2.position) AS max_pos,
				genes.strand,
				t.jn_path,
				t.start_exon,
				t.end_exon,
				t.n_exons
			FROM transcripts t
			LEFT JOIN location loc1 ON t.start_vertex = loc1.location_ID
			LEFT JOIN location loc2 ON t.end_vertex = loc2.location_ID
			LEFT JOIN genes ON t.gene_ID = genes.gene_ID
			WHERE t.transcript_ID IN """ + whitelist_string 
	cursor.execute(query)
	transcript_tuples = cursor.fetchall()

	# Sort based on gene ID
	sorted_transcript_tuples = sorted(transcript_tuples, key=lambda x: x["gene_ID"])	

	gene_groups = {}
	for key,group in itertools.groupby(sorted_transcript_tuples,operator.itemgetter(0)):
		# sort by transcript start position
		gene_groups[key] = sorted(list(group), key=lambda x: x["min_pos"])
	conn.close()

	return gene_groups 

def get_annotations(database, feat_type, whitelist=None):
	""" 
		Extracts annotations from the gene/transcript/exon annotation table of
		the database (depending on choice of feat_type). Limited to rows where
		the annot_name column matches the value of annot.

		Returns:
			annotation_dict: dictionary data structure in which the keys are
							 gene/transcript/exon TALON IDs (depending on 
							 choice of feat_type) and the value is a list of 
							 annotation tuples.
	"""
	# fetch the annotations
	conn = sqlite3.connect(database)
	cursor = conn.cursor()

	table_name = feat_type + "_annotations"

	if whitelist == None:
		query = "SELECT * FROM " + table_name + " WHERE annot_name = '" + annot + \
		 "' OR source = 'TALON'"
	else:
		whitelist_string = "(" + ','.join([str(x) for x in whitelist]) + ")"
		query = "SELECT * FROM " + table_name + " WHERE (annot_name = '" + annot + \
				"' OR source = 'TALON') AND ID IN " + whitelist_string

	cursor.execute(query)
	annotation_tuples = cursor.fetchall()

	# sort based on ID
	sorted_annotations = sorted(annotation_tuples, key=lambda x: x[0]) 

	# group by ID and store in a dictionary
	ID_groups = {}
	for key,group in itertools.groupby(sorted_annotations,operator.itemgetter(0)):
		ID_groups[key] = list(group)

	return ID_groups

def get_annotations(database, feat_type, annot, whitelist = None):
	""" Extracts annotations from the gene/transcript/exon annotation table of
		the database (depending on choice of feat_type). Limited to rows where
		the annot_name column matches the value of annot.

		Returns:
			annotation_dict: dictionary data structure in which the keys are
							 gene/transcript/exon TALON IDs (depending on 
							 choice of feat_type) and the value is a list of 
							 annotation tuples.
	"""
	# Fetch the annotations
	conn = sqlite3.connect(database)
	cursor = conn.cursor()

	table_name = feat_type + "_annotations"

	if whitelist == None:
		query = "SELECT * FROM " + table_name + " WHERE annot_name = '" + annot + \
		 "' OR source = 'TALON'"
	else:
		whitelist_string = "(" + ','.join([str(x) for x in whitelist]) + ")"
		query = "SELECT * FROM " + table_name + " WHERE (annot_name = '" + annot + \
				"' OR source = 'TALON') AND ID IN " + whitelist_string

	cursor.execute(query)
	annotation_tuples = cursor.fetchall()

	# Sort based on ID
	sorted_annotations = sorted(annotation_tuples, key=lambda x: x[0]) 

	# Group by ID and store in a dictionary
	ID_groups = {}
	for key,group in itertools.groupby(sorted_annotations,operator.itemgetter(0)):
		ID_groups[key] = list(group)

	return ID_groups

def get_gene_2_transcripts(database, whitelist):
	""" Creates a dictionary mapping gene IDs to the transcripts that belong to
		them. The columns in each tuple are:
			0: gene ID
			1: transcript ID
			2: chromosome
			3: start position (min of 5' and 3')
			4: end position (max of 5' and 3')
			5: strand
			6: edge path
			7. n_exons
	"""

	conn = sqlite3.connect(database)
	conn.row_factory = sqlite3.Row
	cursor = conn.cursor()
	whitelist_string = "(" + ','.join([str(x) for x in whitelist]) + ")"
	query = """
			SELECT 
				t.gene_ID,
				t.transcript_ID,
				loc1.chromosome,
				MIN(loc1.position,loc2.position) AS min_pos,
				MAX(loc1.position,loc2.position) AS max_pos,
				genes.strand,
				t.jn_path,
				t.start_exon,
				t.end_exon,
				t.n_exons
			FROM transcripts t
			LEFT JOIN location loc1 ON t.start_vertex = loc1.location_ID
			LEFT JOIN location loc2 ON t.end_vertex = loc2.location_ID
			LEFT JOIN genes ON t.gene_ID = genes.gene_ID
			WHERE t.transcript_ID IN """ + whitelist_string 
	cursor.execute(query)
	transcript_tuples = cursor.fetchall()

	# Sort based on gene ID
	sorted_transcript_tuples = sorted(transcript_tuples, key=lambda x: x["gene_ID"])	

	gene_groups = {}
	for key,group in itertools.groupby(sorted_transcript_tuples,operator.itemgetter(0)):
		# Sort by transcript start position
		gene_groups[key] = sorted(list(group), key=lambda x: x["min_pos"])
	conn.close()

	return gene_groups 

def fetch_exon_locations(database):
	""" Queries the database to create a dictionary mapping exon IDs to 
		the chromosome, start, end, and strand of the exon """

	conn = sqlite3.connect(database)
	cursor = conn.cursor()

	query = """
			SELECT 
				e.edge_ID,
				loc1.chromosome,
				MIN(loc1.position,loc2.position),
				MAX(loc1.position,loc2.position),
				e.strand
			 FROM edge e
			 LEFT JOIN location loc1 ON e.v1 = loc1.location_ID
			 LEFT JOIN location loc2 ON e.v2 = loc2.location_ID
			 WHERE e.edge_type = 'exon';"""

	cursor.execute(query)
	exon_location_tuples = cursor.fetchall()

	# Create dictionary
	exon_locations = {}
	for loc_tuple in exon_location_tuples:
		exon_ID = loc_tuple[0]
		exon_locations[exon_ID] = loc_tuple[1:]

	conn.close() 
	return exon_locations

def check_annot_validity(annot, database):
    """ Make sure that the user has entered a correct annotation name """

    conn = sqlite3.connect(database)
    cursor = conn.cursor()

    cursor.execute("SELECT DISTINCT annot_name FROM gene_annotations")
    annotations = [str(x[0]) for x in cursor.fetchall()]
    conn.close()

    if "TALON" in annotations:
        annotations.remove("TALON")

    if annot == None:
        if len(annotations) == 1:
            annot = annotations[0]

        else:
            message = "Please provide a valid annotation name. " + \
                  "In this database, your options are: " + \
                  ", ".join(annotations)
            raise Exception(message)

    if annot not in annotations:
        message = "Annotation name '" + annot + \
                  "' not found in this database. Try one of the following: " + \
                  ", ".join(annotations)
        raise Exception(message)

    return annot


import pytest
import sys
import numpy as np
import swan_vis as swan

# i suppose this is also testing the df init because 
# everything needs to be here? And in the right order -\(' ')/-
class TestMergeSGs(object):
	def test_merge_sgs(self):
		a_gtf = 'input_files/annot.gtf'
		b_gtf = 'input_files/annot_2.gtf'
		sg = swan.SwanGraph()
		sg.add_dataset('a', a_gtf, include_isms=True)
		sg.add_dataset('b', b_gtf, include_isms=True)

		print(sg.loc_df.head())
		print(sg.edge_df.head())
		print(sg.t_df.head())

		# print(sg.loc_df[['chrom', 'coord', 'strand', 'a', 'b']])
		# print(sg.edge_df[['edge_type', 'a', 'b']])
		# print(sg.t_df[['path', 'a', 'b']])

		# check that the format of dfs are ok
		assert sg.loc_df.index.names == ['vertex_id']
		control = ['coord', 'chrom', 'strand', 'a', 'b', 'vertex_id',
				   'internal', 'TSS', 'TES'] 
		test = sg.loc_df.columns.tolist()
		check_pairs(control, test)

		assert sg.edge_df.index.names == ['edge_id']
		control = ['v1', 'v2', 'edge_type', 'strand', 'a', 'b', 'edge_id'] 
		test = sg.edge_df.columns.tolist()
		check_pairs(control, test)

		assert sg.t_df.index.names == ['tid']
		control = ['tid', 'gid', 'gname', 'path', 'a', 'b']
		test = sg.t_df.columns.tolist()
		check_pairs(control, test)

		# test that loc_df merging happened correctly
		# query chr, coord, strand, a and b columns
		chrs = sg.loc_df['chrom'].tolist()
		control = [1,1,1,1,1,1,1,7,7,7,7,1,1,1,1,1,4,4]
		print('test chrs: ')
		print(chrs)
		print('control chrs: ')
		print(chrs)
		assert control == control

		coords = sg.loc_df['coord'].tolist()
		control = [1,90,100,500,600,900,1000,1,10,15,
				   20,2000,1500,1000,900,800,4000,1000]
		print('test coords: ')
		print(coords)
		print('control coords: ')
		print(control)
		assert coords == control

		strand = sg.loc_df['strand'].tolist()
		control = ['+','+','+','+','+','+','+','+','+',
				   '+','+','-','-','-','-','-','-','-']
		print('test strands: ')
		print(strand)
		print('control strands: ')
		print(control)
		assert strand == control

		a = sg.loc_df['a'].tolist()
		control = [bool(i) for i in [1,0,1,1,1,1,1,0,0,0,0,1,1,1,1,0,1,1]]
		print('test a presence: ')
		print(a)
		print('control a presence: ')
		print(control)
		assert a == control

		b = sg.loc_df['b'].tolist()
		control = [bool(i) for i in [1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,0,0]]
		print('test b presence: ')
		print(b)
		print('control b presence: ')
		print(control)
		assert b == control

		# test that edge_df merging and id mapping happened correctly
		# query edge_id, edge_type, a and b columns
		edge_id = sg.edge_df['edge_id'].tolist()
		control = [(0,1),(0,2),(1,3),(2,3),(3,4),(4,5),(5,6),
				   (7,8),(8,9),(9,10),(11,12),(12,13),(13,14),(13,15),(16,17)]
		print('test edge_ids: ')
		print(edge_id)
		print('control edge_ids')
		print(control)
		assert edge_id == control

		edge_type = sg.edge_df['edge_type'].tolist()
		control = ['exon','exon','intron','intron','exon','intron','exon',
				   'exon','intron','exon','exon','intron','exon','exon','exon']
		print('test edge_types: ')
		print(edge_type)
		print('control edge_types: ')
		print(control)
		assert edge_type == control

		a = sg.edge_df['a'].tolist()
		control = [bool(i) for i in [0,1,0,1,1,1,1,0,0,0,1,1,1,0,1]]
		print('test a presence: ')
		print(a)
		print('control a presence: ')
		print(control)
		assert a == control

		b = sg.edge_df['b'].tolist()
		control = [bool(i) for i in [1,1,1,1,1,1,1,1,1,1,1,1,0,1,0]]
		print('test b presence: ')
		print(b)
		print('control b presence: ')
		print(control)
		assert b == control

		# test that t_df merging and id mapping happened correctly
		# query tid, path, a and b columns
		tid = sg.t_df['tid'].tolist()
		control = ['ENST01','ENST02','ENST03','ENST04','ENST07','ENST08']
		print('test tids: ')
		print(tid)
		print('control tids: ')
		print(control)
		assert tid == control

		paths = [tuple(path) for path in sg.t_df['path'].tolist()]
		control = [(0,2,3,4,5,6),(0,1,3,4,5,6),(11,12,13,14),(11,12,13,15),
				   (16,17),(7,8,9,10)]
		print('test paths: ')
		print(paths)
		print('control paths: ')
		print(control)
		assert paths == control

		a = sg.t_df['a'].tolist()
		control = [bool(i) for i in [1,0,1,0,1,0]]
		print('test a presence: ')
		print(a)
		print('control a presence: ')
		print(control)
		assert a == control

		b = sg.t_df['b'].tolist()
		control = [bool(i) for i in [1,1,0,1,0,1]]
		print('test b presence: ')
		print(b)
		print('control b presence: ')
		print(control)
		assert b == control

	# test novelty merging
	def test_novelty_merging(self):
		a_gtf = 'input_files/annot.gtf'
		b_gtf = 'input_files/annot_2.gtf'

		# only one dataset has novelty categorizations
		# and there are transcripts in the dataset that
		# doesn't have categories that aren't in the
		# dataset that does
		print('Testing if novelty merging works...')
		sg = swan.SwanGraph()
		sg.add_dataset('a', a_gtf, include_isms=True)
		sg.t_df['novelty'] = ['Known', 'NIC', 'ISM']
		sg.add_dataset('b', b_gtf, include_isms=True)
		test = sg.t_df.apply(lambda x: (x.tid, x.novelty), axis=1)
		control = [('ENST01', 'Known'), ('ENST02', 'Undefined'), ('ENST08', 'Undefined'),
				   ('ENST03', 'NIC'), ('ENST04', 'Undefined'), ('ENST07', 'ISM')]
		check_pairs(control, test)

		# one dataset has already been given "undefined" label 
		# that needs to be overwritten
		print('Testing merging with novelty=undefined given a new label')
		sg_b = swan.SwanGraph()
		sg_b.add_dataset('b_2', b_gtf, include_isms=True)
		sg_b.t_df['novelty'] = ['ISM', 'Antisense', 'Intergenic', 'Genomic']

		print(sg.t_df)
		print(sg_b.t_df)
		sg.merge_dfs(sg_b, 'b_2')
		control = [('ENST01', 'Ambiguous'), ('ENST02', 'Antisense'), ('ENST08', 'Genomic'),
				   ('ENST03', 'NIC'), ('ENST04', 'Intergenic'), ('ENST07', 'ISM')]
		test = sg.t_df.apply(lambda x: (x.tid, x.novelty), axis=1)
		check_pairs(control, test)


		# both datasets have novelty categorizations
		ab_gtf = 'input_files/annot_3.gtf'
		sg_a = swan.SwanGraph()
		sg_a.add_dataset('a', ab_gtf, include_isms=True)
		sg_a.t_df['novelty'] = ['Known', 'Known', 'ISM',
								np.nan, 'NIC', 'NNC']
		sg_b = swan.SwanGraph()
		sg_b.add_dataset('b', ab_gtf, include_isms=True)
		sg_b.t_df['novelty'] = ['Known', 'ISM', np.nan,
								'NIC', 'NNC', 'NNC']
		sg_a.merge_dfs(sg_b, 'b')
		test = sg_a.t_df.apply(lambda x: (x.tid, x.novelty), axis=1)
		control = [('ENST01', 'Known'), ('ENST02', 'Ambiguous'), ('ENST08', 'NNC'),
				   ('ENST03', 'ISM'), ('ENST04', 'NIC'), ('ENST07', 'Ambiguous')]
		check_pairs(control, test)

		# neither dataset has novelty types
		sg = swan.SwanGraph()
		sg.add_dataset('a', a_gtf, include_isms=True)
		sg.add_dataset('b', b_gtf, include_isms=True)
		assert 'novelty' not in sg.t_df.columns

def check_pairs(control, test):
	print('control')
	print(control)
	print('test')
	print(test)
	for t in test:
		assert t in control
	assert len(test) == len(control)









import networkx as nx
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os
import copy
from collections import defaultdict
import matplotlib.patches as pch
from swan_vis.utils import *
from swan_vis.graph import *
import time

class PlottedGraph(Graph):

	def __init__(self):
		self.indicate_dataset = False
		self.indicate_novel = False
		self.browser = False 
		self.gid = None
		self.tid = None
		self.dataset_cols = []

		# plotting colors
		gray = '#DCDCDC'
		yellow = '#F0E442'
		light_blue = '#56B4E9'
		orange = '#E69F00'
		pink = '#CC79A7'
		green = '#009E73'

		gray_light_blue = '#c1ddec'
		gray_yellow = '#f8f6db'
		gray_orange = '#ebdec3'
		gray_pink = '#efdae5'
		gray_green = '#cdf5ea'

		self.color_dict = {'intron': {'normal': pink, 'gray': gray_pink},
					  'exon': {'normal': green, 'gray': gray_green},
					  'TSS': {'normal': light_blue, 'gray': gray_light_blue},
					  'TES': {'normal': orange, 'gray': gray_orange},
					  'internal': {'normal': yellow, 'gray': gray_yellow}}

	# initialize what needs to be from a previous pg 
	# or otherwise
	def init_plot_settings(self, sg,
						   gid=None, 
						   tid=None,
						   indicate_dataset=False,
						   indicate_novel=False,
						   browser=False):

		# save some old stuff
		old_indicate_dataset = self.indicate_dataset
		old_indicate_novel = self.indicate_novel
		old_gid = self.gid
		old_tid = self.tid
		old_browser = self.browser
		old_dataset_cols = self.dataset_cols

		# init some stuff
		self.indicate_dataset = indicate_dataset 
		self.indicate_novel = indicate_novel
		self.gid = gid 
		self.tid = tid
		self.browser = browser
		self.dataset_cols = sg.get_dataset_cols()

		# more human-readable graph types
		if self.tid:
			self.graph_type = 'transcript_path'
		else:
			self.graph_type = 'summary'


		# if we have new datasets
		if len(list(set(old_dataset_cols)-set(self.dataset_cols))) != 0:
			self.new_gene(sg)
			if not self.browser:
				self.calc_node_edge_styles()
				self.get_ordered_edges()

		# plotting the same transcript
		elif old_tid == self.tid and self.tid != None:
			if not browser: 
				if old_indicate_dataset != self.indicate_dataset \
				or old_indicate_novel != self.indicate_novel:
					self.calc_node_edge_styles()

		# plotting a different transcript
		elif old_tid != self.tid and self.tid != None:

			self.gid = sg.get_gid_from_tid(self.tid)
			if old_gid != self.gid:
				self.new_gene(sg)
			else:
				self.path = self.get_path_from_tid(self.tid)
			if not self.browser:
				self.calc_node_edge_styles()
				self.get_ordered_edges()

		# plotting a different gene
		elif old_gid != self.gid:
			self.new_gene(sg)
			if not self.browser:
				self.calc_node_edge_styles()

		elif old_indicate_dataset != self.indicate_dataset \
		or old_indicate_novel != self.indicate_novel:
			self.calc_node_edge_styles()

	# update pg to contain information for plotting a new gene
	def new_gene(self, sg):
		sg = subset_on_gene(sg, self.gid)
		self.loc_df = sg.loc_df
		self.edge_df = sg.edge_df
		self.t_df = sg.t_df
		self.G = sg.G

		self.g_min, self.g_max = self.get_gene_min_max(self.gid)
		self.strand = self.get_strand_from_gid(self.gid)

		if self.tid:
			self.path = self.get_path_from_tid(self.tid)

		if not self.browser:
			self.ordered_nodes = self.get_ordered_nodes()
			self.calc_pos_sizes()
			self.calc_edge_curves()

	# settings that can be initialized if we're plotting a new transcript
	# (or a new gene)
	def calc_node_edge_styles(self):
		self.calc_node_colors()
		self.calc_edge_colors()
		self.calc_edge_linestyles()

	###############################################################################
	################### Getting plotting settings for nodes/edges #################
	###############################################################################
		
	# calculates the positions and sizes of edges/nodes based on the 
	# number of nodes in the graph
	def calc_pos_sizes(self):

		pos = defaultdict()
		ordered_nodes = self.ordered_nodes

		y_coord = 0
		x_coords = np.linspace(-1, 1, len(ordered_nodes))
		for x_coord, n in zip(x_coords, ordered_nodes): 
			pos[n] = [x_coord, y_coord]

		x = len(ordered_nodes)

		# area-related sizes need to be non-linearly scaled
		# calculated by fitting power series curve to handpicked sizes
		node_size = 19248*(x**-1.14) 
		label_size = 43.9*(x**-0.484)

		# linearly-related sizes
		edge_width = -x/18 + (121/18)
		if edge_width < 1:
			edge_width = 1

		# assign fields to plotted graph object 
		self.pos = pos
		self.node_size = node_size
		self.label_size = label_size
		self.rad_scale = 0.32
		self.edge_width = edge_width

	# calculate the color for each node
	def calc_node_colors(self):
		self.loc_df = self.loc_df.apply(
			lambda x: self.get_node_color(x), axis=1)

	# does the node satisfy indicate_novel or indicate_dataset
	# plotting settings?
	def is_novel_or_in_dataset(self, x):
		if self.indicate_novel and is_novel(x):
			if self.graph_type == 'transcript_path' \
			and x.vertex_id in self.path:
				return 'k'
			elif self.graph_type == 'summary':
				return 'k'
			else:
				return "#999999"
		elif self.indicate_dataset and in_dataset(self.indicate_dataset, x):
			if self.graph_type == 'transcript_path' \
			and x.vertex_id in self.path:
				return 'k'
			elif self.graph_type == 'summary':
				return 'k'
			else:
				return "#999999"
		return None

	# get the node color 
	def get_node_color(self, x):

		# transcript path through summary graph
		if self.graph_type == 'transcript_path':

			# vertices in the path should be colored by their roles in 
			# the plotted transcript
			if x.vertex_id == self.path[0]:
				x['color'] = self.color_dict['TSS']['normal']
			elif x.vertex_id == self.path[-1]:
				x['color'] = self.color_dict['TES']['normal']
			elif x.vertex_id in self.path:
				x['color'] = self.color_dict['internal']['normal']
			
			# vertices not in the path should be less colored
			# and colored according to their "most unique"
			# role in the current gene ie
			# internal < TSS < TES			
			else:
				if x.internal: color = self.color_dict['internal']['gray']
				if x.TSS: color = self.color_dict['TSS']['gray']
				if x.TES: color = self.color_dict['TES']['gray']
				x['color'] = color

		# gene summary graph
		else:

			# vertices in the summary graph should be colored 
			# according to their "most unique" 
			# role in the gene ie
			# internal < TSS < TES	
			if x.internal: color = self.color_dict['internal']['normal']
			if x.TSS: color = self.color_dict['TSS']['normal']
			if x.TES: color = self.color_dict['TES']['normal']
			x['color'] = color

		# if the node needs to be outlined due to indicate_dataset
		# or indicate_novel
		if self.indicate_novel or self.indicate_dataset:
			x['edgecolor'] = self.is_novel_or_in_dataset(x)
		else:
			x['edgecolor'] = None

		return x

	# get edge curve settings
	def calc_edge_curves(self):
		edge_dict = {'pos': 'arc3,rad=',
					 'neg': 'arc3,rad=-',
					 'straight': None}
		ordered_edges = [(n,m) for n,m in zip(self.ordered_nodes[:-1],
			self.ordered_nodes[1:])]
		self.edge_df['raw_index'] = [i for i in range(len(self.edge_df.index))]
		self.edge_df['curve'] = self.edge_df.apply(
			lambda x: self.get_edge_curve(x, ordered_edges, edge_dict), axis=1)
		self.edge_df.drop('raw_index', axis=1, inplace=True)

	# get the curve/style of the edge
	def get_edge_curve(self, x, ordered_edges, edge_dict):

		# over 20 nodes, all should be curved
		if len(ordered_edges) < 20:
			if x.edge_id in ordered_edges:
				return edge_dict['straight']

		# make the arcs pretty
		dist = self.pos[x.v2][0] - self.pos[x.v1][0]
		rad = self.rad_scale/dist

		if x.raw_index % 2 == 0:
			return edge_dict['pos']+str(rad)
		else:
			return edge_dict['neg']+str(rad)

	# get the colors of all the edges
	def calc_edge_colors(self):
		self.edge_df['color'] = self.edge_df.apply(
			lambda x: self.get_edge_color(x), axis=1)

	# get the color of the edge
	def get_edge_color(self, x):
		# firstly, if we're given a path, 
		# only color the edges that are in the path
		color_dict = self.color_dict
		if self.tid:
			path_edges = [(self.path[i],self.path[i+1])
						   for i in range(len(self.path)-1)]
			if x.edge_id in path_edges:
				color = color_dict[x.edge_type]['normal']
			else:
				color = color_dict[x.edge_type]['gray']
		# otherwise just color them all
		else:
			color = color_dict[x.edge_type]['normal']
		return color

	# get the styles of the edges (dashed or not)
	def calc_edge_linestyles(self):
		self.edge_df['line'] = self.edge_df.apply(
			lambda x: self.get_edge_linestyle(x),
			axis=1)

	# get the style of the line for an edge
	def get_edge_linestyle(self, x):
		style = None
		# dashed if novel or unique to dataset
		# if is_novel(self.indicate_novel, x) or unique_to_dataset(self.indicate_dataset, x, dataset_cols):
		if self.indicate_novel:
			if is_novel(x):
				style = 'dashed'
		elif self.indicate_dataset:
			if in_dataset(self.indicate_dataset, x):
				style = 'dashed'
		return style

	###############################################################################
	########################## Actually plotting stuff ############################
	###############################################################################

	# plots input plotted graph according to user's choices
	def plot_graph(self):

		# swan graph
		if not self.browser:
			self.plot_swan_graph()
		# browser track
		else:
			self.plot_browser_path()

	###############################################################################
	############################ Swan graph plotting ##############################
	###############################################################################

	# plots swan graph of current plotted graph
	def plot_swan_graph(self):

		# plotting stuff
		plt.figure(1, figsize=(14,2.8), frameon=False)
		plt.xlim(-1.05, 1.05)
		plt.ylim(-1.05, 1.05) 

		# plot edges, nodes, and labels
		self.plot_edges()

		# relabel nodes with strings for the purpose of labelling them
		relabel_map = {k:(k if type(k) == str else int(k)) for k in self.G.nodes}
		self.G = nx.relabel_nodes(self.G, relabel_map)
		nx.draw_networkx_labels(self.G,
			self.pos,
			font_size=self.label_size)

		self.plot_nodes()

		# remove axis
		plt.axis('off')

	# plots edges from edge_df
	def plot_edges(self):
		for _, entry in self.edge_df.iterrows():
			edge = entry.edge_id
			nx.draw_networkx_edges(self.G, self.pos,
				edgelist=[edge],
				width=self.edge_width,
				edge_color=entry.color,
				connectionstyle=entry.curve,
				style=entry.line)

	# plots nodes from loc_df
	def plot_nodes(self):
		for _, entry in self.loc_df.iterrows():
			node = entry.vertex_id
			nx.draw_networkx_nodes(self.G, self.pos,
				nodelist=[node],
				node_color=entry.color,
				node_size=self.node_size,
				edgecolors=entry.edgecolor)

	###############################################################################
	######################## Browser track style plotting #########################
	###############################################################################

	# plots the browser track representation of the transcript from self.path
	def plot_browser_path(self):

		# which gene does this transcript come from?
		# what are the min/max coords in the gene?
		gid = self.gid
		g_min = self.g_min
		g_max = self.g_max
		t_min, t_max = self.get_transcript_min_max(self.tid)
		g_len = g_max - g_min

		x_min = int(g_min-(6/(g_max-g_min)))
		x_max = int(g_max+(6/(g_max-g_min)))

		strand = self.strand
		path = self.path

		# plotting init
		plt.figure(1, figsize=(14,2.8), frameon=False)
		# plt.xlim(x_min, x_max)
		plt.xlim(g_min, g_max)
		plt.ylim(-1.05, 1.05)
		ax = plt.gca()
		teal = '#014753'

		# plot each exon as a rectangle
		y_coord = -0.1
		height = 0.2
		exons = [(v1,v2) for v1,v2 in zip(path[:-1],path[1:])][::2]
		introns = [(v1,v2) for v1,v2 in zip(path[:-1],path[1:])][1::2]
		for v1,v2 in exons:
			x_coord = self.loc_df.loc[v1, 'coord']
			width = self.loc_df.loc[v2, 'coord'] - x_coord
			rect = pch.Rectangle((x_coord,y_coord), width, height, color=teal)
			ax.add_patch(rect)

		# plot each intron as a line
		dist = 0.001*g_len
		plt.plot([t_min+dist, t_max-dist], [0,0], color=teal)

		# reverse plot if we're on the minus strand
		if strand == '-':
			plt.gca().invert_xaxis()

		# remove axis
		plt.axis('off')

		def get_tick_coords(loc_df, gene_len, exons, g_min, g_max, strand):

			tick_coords = list(np.linspace(g_min, g_max, 40))

			# remove ticks that are before the start of the first exon
			start = loc_df.loc[exons[0][1], 'coord']
			if strand == '+':
				tick_coords = [t for t in tick_coords if t > start]
			else:
				tick_coords = [t for t in tick_coords if t < start]

			# remove ticks that are after the end of the last exon
			stop = loc_df.loc[exons[-1][1], 'coord']
			if strand == '+':
				tick_coords = [t for t in tick_coords if t < stop]
			else:
				tick_coords = [t for t in tick_coords if t > stop]

			# remove ticks in and around the area of plotted exons
			dist = 0.001*gene_len
			for v1,v2 in exons:
				start = loc_df.loc[v1, 'coord']
				stop = loc_df.loc[v2, 'coord']
				tick_coords = [t for t in tick_coords
							   if t < start-dist or t > stop+dist]

			# if we only have one intron, and nothing made the cut, just stick
			# a tick in the middle of the intron
			if len(exons) == 2 and not tick_coords:
				tick_coords = [(loc_df.loc[exons[0][1],'coord']+loc_df.loc[exons[1][0],'coord'])/2]

			return tick_coords 

		# get coordinates for evenly-spaced ticks indicating strandedness
		# ala genome browser
		tick_coords = get_tick_coords(self.loc_df, g_len, exons,
			g_min, g_max, self.strand)
		plt.plot(tick_coords, [0 for i in range(len(tick_coords))],
			color=teal, marker='4')

	# plots the scale of a browser style graph
	def plot_browser_scale(self):

		# get plot limits
		g_min = self.g_min
		g_max = self.g_max 

		plt.figure(1, figsize=(14,1), frameon=False)
		plt.xlim(g_min, g_max)
		plt.ylim(-1.05, 1.05)
		ax = plt.gca()

		y_coord = -0.5
		height = 1
		gene_len = g_max - g_min
		width = int((10/10000)*gene_len)

		# try different scale lengths until we've reached something optimal
		scale_len = 500
		while 1:
			if scale_len >= gene_len:
				break
			scale_len *= 10

		# dial it back so we show a nice chunk instead of the scale of the entire gene
		scale_len /= 10

		# are we going to indicate this in bp or kb?
		if scale_len < 1000: 
			scale_unit = 'bp'
			text_scale = scale_len
		elif scale_len < 10**6:
			scale_unit = 'kb'
			text_scale =  scale_len / 1000
		else:
			scale_unit = 'mb'
			text_scale = scale_len / 10**6
		text_scale = int(text_scale)

		# center the edges of the scale bar
		gene_mid_pt = (g_max+g_min)/2
		scale_start = gene_mid_pt - (scale_len/2)
		scale_end = gene_mid_pt + (scale_len/2)

		# start bar
		rect = pch.Rectangle((scale_start, y_coord),
			width=width,
			height=height,
			color='k')
		ax.add_patch(rect)

		# end bar
		rect = pch.Rectangle((scale_end, y_coord),
			width=width,
			height=height,
			color='k')
		ax.add_patch(rect)

		# plot the line in between the bars
		plt.plot([scale_start+width, scale_end-width], [0,0],
			color='k',
			linewidth=2)

		# add scale text
		x_coord = scale_start - (gene_len/9)
		y_coord = -0.2
		txt = '{} {}'.format(text_scale, scale_unit)
		ax.text(x_coord, y_coord, txt, fontsize=30)

	# gets nodes ordered by genomic position 
	def get_ordered_nodes(self):
		loc_ids = self.loc_df.vertex_id.tolist()
		coords = self.loc_df.coord.tolist()
		ordered_nodes = [i[0] for i in sorted(zip(loc_ids, coords),
			key=lambda x: x[1])]

		# check if forward or reverse strand 
		if self.strand == '-':
			ordered_nodes.reverse()
		return ordered_nodes

	# orders edges by those present in the transcript and those not present in the transcript
	def get_ordered_edges(self):
		path_edges = [(self.path[i],self.path[i+1])
			   for i in range(len(self.path)-1)]
		self.edge_df['in_transcript'] = self.edge_df.apply(lambda x:
			1 if x.edge_id in path_edges else 0, axis=1)
		self.edge_df.sort_values(by='in_transcript', inplace=True)
		self.edge_df.drop('in_transcript', axis=1, inplace=True)

# is this entry unique from all other datasets?
# and in the provided dataset?
def unique_to_dataset(dataset_name, x, dataset_cols):
	if not dataset_name: return False
	if x[dataset_name] and not any(x[dataset_cols].tolist()):
		return True
	return False

# is this entry in the provided dataset?
def in_dataset(dataset_name, x):
	return x[dataset_name]

# is this entry annotated?
def is_novel(x):
	if x.annotation: return False
	return True






import networkx as nx
import numpy as np
import pandas as pd
import copy
from swan_vis.utils import *

# super class that both SwanGraph and PlottedGraph inherit from.
# all functions that both subclasses will use are in here.
class Graph:
	def __init__(self):
		""" 
		A general graph class to represent a transcriptome.

		Attributes
		----------
		datasets (list of str):
			Names of datasets in the Graph
		counts (list of str):
			Names of columns holding counts in the Graph
		tpm (list of str):
			Names of columns holding tpm values in the Graph
		loc_df (pandas DataFrame): 
			DataFrame of all unique observed genomic 
			coordinates in the transcriptome
		edge_df (pandas DataFrame):
			DataFrame of all unique observed exonic or intronic
			combinations of splice sites in the transcriptome
		t_df (pandas DataFrame): 
			DataFrame of all unique transcripts found 
			in the transcriptome
		"""

		self.datasets = []
		self.counts = []
		self.tpm = []
		
		self.loc_df = pd.DataFrame(columns=['chrom', 'coord',
									   'strand','vertex_id',
									   'TSS',
									   'TES',
									   'internal'])
		self.edge_df = pd.DataFrame(columns=['edge_id', 'edge_type',
									    'strand', 'v1', 'v2'])
		self.t_df = pd.DataFrame(columns=['tid', 'gid',
									 'gname', 'path'])

	##########################################################################
	################# Related to checking contents of Graph ##################
	##########################################################################

	# check that input datasets are in the Graph:
	def check_datasets(self, datasets):

		# make sure we have an iterable
		if type(datasets) != list:
			datasets = [datasets]

		g_datasets = self.get_dataset_cols()
		for d in datasets:
			if d not in g_datasets:
				raise Exception('Dataset {} not present in graph. '
								'Datasets in graph are {}'.format(d, g_datasets))


	# check that input datasets have abundance data in the Graph:
	def check_abundances(self, datasets):

		# make sure we have an iterable
		if type(datasets) != list:
			datasets = [datasets]

		ab_cols = self.get_count_cols()
		for d in datasets:
			if '{}_counts'.format(d) not in ab_cols:
				raise Exception('Abundance for dataset {} not present in graph. '
								'Datasets with abundance information '
								'in graph are {}'.format(d, ab_cols))

	# check that de tests have been run 
	def check_de(self, de_type):
		if de_type == 'transcript':
			if not self.det_test.empty:
				return True
		elif de_type == 'gene':
			if not self.deg_test.emtpy:
				return False
		else:
			return False

	# check if there are even any datasets
	def check_if_any_datasets(self, task):
		if self.datasets == None:
			raise Exception('No datasets found in graph. '
				'Cannot perform {}'.format(task))

	# check if gid is in SwanGraph
	def check_gene(self, gid):
		if gid not in self.t_df.gid.tolist():
			raise Exception('Gene {} not found in Graph.'.format(gid))

	# check if tid is in SwanGraph
	def check_transcript(self, tid):
		if tid not in self.t_df.tid.tolist():
			raise Exception('Transcript {} not found in Graph.'.format(tid))


	##########################################################################
	####################### Related to creating Graph ########################
	##########################################################################

	# update ids according to coordinates in loc_df, edge_df, and t_df
	def update_ids(self, id_map=None):

		# if the user didn't already input an id map, just order
		# according to genomic position
		if not id_map:
			id_map = self.get_ordered_id_map()

		# convert dfs into dicts for the next steps
		self.dfs_to_dicts()

		# update the ids according to id_map
		self.update_loc_df_ids(id_map)
		self.update_edge_df_ids(id_map)
		self.update_t_df_paths(id_map)

		# convert back to dfs
		self.dicts_to_dfs()	

	# get a dictionary mapping vertex id to ordered new vertex id
	def get_ordered_id_map(self):

		# split loc_df into + and - strand parts
		plus_loc_df = self.loc_df.loc[self.loc_df.strand == '+'].copy(deep=True)
		minus_loc_df = self.loc_df.loc[self.loc_df.strand == '-'].copy(deep=True)

		# sort each of the dfs by chrom, coord either ascending
		# or descending based on strand
		plus_loc_df.sort_values(['chrom', 'coord'],
								 ascending=[True, True],
								 inplace=True)
		minus_loc_df.sort_values(['chrom', 'coord'],
								  ascending=[True, False],
								  inplace=True)

		# concatenate the two dfs
		self.loc_df = pd.concat([plus_loc_df, minus_loc_df])

		# dictionary mapping vertex_id to new_id
		self.loc_df['new_id'] = [i for i in range(len(self.loc_df.index))]

		# account for combined nodes, will still be labelled as c#,
		# but can reorder the '#' according to starting coord
		if 'combined' in self.loc_df.columns:
			combined_ids = self.loc_df.loc[self.loc_df.combined, 'vertex_id']
			combined_new_ids = ['c{}'.format(i) for i in range(len(combined_ids))]
			self.loc_df.loc[combined_ids, 'new_id'] = combined_new_ids

		id_map = self.loc_df['new_id'].to_dict()
		self.loc_df.drop('new_id', axis=1, inplace=True)

		return id_map

	# update vertex ids in loc_df
	def update_loc_df_ids(self, id_map):

		loc_df = {}
		for old_id, new_id in id_map.items():
			loc_df[new_id] = self.loc_df[old_id]
			loc_df[new_id]['vertex_id'] = new_id
		self.loc_df = loc_df

	# update vertex ids in edge_df
	def update_edge_df_ids(self, id_map):

		edge_df = {}
		for edge_id, item in self.edge_df.items():
			new_id = (id_map[edge_id[0]],id_map[edge_id[1]])
			edge_df[new_id] = item
			edge_df[new_id]['v1'] = new_id[0]
			edge_df[new_id]['v2'] = new_id[1]
			edge_df[new_id]['edge_id'] = new_id
		self.edge_df = edge_df

	# update vertex ids in t_df
	def update_t_df_paths(self, id_map):

		t_df = {}
		for tid, item in self.t_df.items():
			path = item['path']
			new_path = []
			for n in path:
				new_path.append(id_map[n])
			t_df[tid] = item
			t_df[tid]['path'] = new_path
		self.t_df = t_df

	# create the graph object from the dataframes
	def create_graph_from_dfs(self):

		G = nx.DiGraph()

		# add nodes to graph from transcript paths
		paths = self.t_df.path.tolist()
		for path in paths:
			nx.add_path(G, path)

		self.G = G

	# order edge df based on source id
	def order_edge_df(self):
		self.edge_df.sort_values(by=['v1', 'v2'], inplace=True)


	##########################################################################
	####### Functions to switch back and forth between dfs and dicts #########
	##########################################################################

	# convert loc_df, edge_df, and t_df to dictionaries
	def dfs_to_dicts(self):

		self.loc_df = self.loc_df.to_dict('index')
		self.edge_df = self.edge_df.to_dict('index')
		self.t_df = self.t_df.to_dict('index')

	# convert dictionary versions of loc_df, edge_df, and t_df to dfs
	def dicts_to_dfs(self):

		# loc_df
		self.loc_df = pd.DataFrame.from_dict(self.loc_df, orient='index')
		self.loc_df.index.names = ['vertex_id']

		# pandas interprets the tuple as a multiindex so we need to fix it
		self.edge_df = pd.DataFrame.from_dict(self.edge_df, orient='index')
		self.edge_df.reset_index(drop=True, inplace=True)
		self.edge_df = create_dupe_index(self.edge_df, 'edge_id')
		self.edge_df = set_dupe_index(self.edge_df, 'edge_id')

		# t_df
		self.t_df = pd.DataFrame.from_dict(self.t_df, orient='index')
		self.t_df.index.names = ['tid']

	##########################################################################
	############################# Other utilities ############################
	##########################################################################

	# check if anything has been added to the graph yet
	def is_empty(self):
		if len(self.datasets) == 0: 
			return True
		else: 
			return False

	# check if novelty information is in the graph
	def has_novelty(self):
		if 'novelty' in self.t_df.columns.tolist():
			return True
		else: return False

	# gets the names of the dataset columns in the graph
	def get_dataset_cols(self, include_annotation=True):
		if include_annotation:
			return self.datasets
		elif not include_annotation:
			if 'annotation' not in self.datasets:
				return self.datasets
			else: 
				datasets = copy.deepcopy(self.datasets)
				datasets.remove('annotation')
				return datasets

	# gets the names of the counts columns in the graph
	# returns None if no counts have been added
	# if datasets option given, returns the counts 
	# columns associated with the input datasets
	def get_count_cols(self, datasets=None):

		if datasets:
			if type(datasets) != list:
				datasets = [datasets]
			self.check_abundances(datasets)
			counts_cols = []
			for d in datasets:
				counts_cols.append('{}_counts'.format(d))
			return counts_cols

		return self.counts

	# gets the names of tpm columns in the graph
	# returns None if no counts have been added
	# if datasets option given, returns the counts 
	# columns associated with the input datasets
	def get_tpm_cols(self, datasets=None):

		if datasets:
			if type(datasets) != list:
				datasets = [datasets]
			self.check_abundances(datasets)
			tpm_cols = []
			for d in datasets:
				tpm_cols.append('{}_tpm'.format(d))
			return tpm_cols

		return self.tpm

	# gets strandedness of transcript from transcript id
	def get_strand_from_tid(self, tid):
		return self.loc_df.loc[self.t_df.loc[tid, 'path'][0], 'strand']

	# gets strandedness of transcript from gene id
	def get_strand_from_gid(self, gid):
		vertex = self.t_df.loc[self.t_df.gid == gid].path.tolist()[0][0]
		return self.loc_df.loc[vertex, 'strand']

	# get the path from the transcript id
	def get_path_from_tid(self, tid):
		return self.t_df.loc[tid].path

	# get the gene id from the gene name
	def get_gid_from_gname(self, gname):
		try:
			gid = self.t_df.loc[self.t_df.gname == gname, 'gid'].tolist()[0]
		except:
			gid = gname
		return gid

	# get the gene id from the transcript id
	def get_gid_from_tid(self, tid):
		return self.t_df.loc[tid, 'gid']

	# returns the min and max coordinates of an input gene
	def get_gene_min_max(self, gid):

		# get all starts and stops from gene and their coordinates
		paths = self.t_df.loc[self.t_df.gid == gid].path.tolist()
		starts = np.unique([path[0] for path in paths]).tolist()
		stops = np.unique([path[-1] for path in paths]).tolist()
		coords = self.loc_df.loc[starts+stops, 'coord']

		return int(min(coords)), int(max(coords))

	# returns the min and max coordinates of an input transcript
	def get_transcript_min_max(self, tid):

		path = self.t_df.loc[tid, 'path']
		ends = [path[0], path[-1]]
		end_coords = self.loc_df.loc[ends, 'coord']

		return int(min(end_coords)), int(max(end_coords))

###########################################################################
################################## Extras #################################
##########################################################################

 # subset a graph based on a gene
def subset_on_gene(sg, gid):

	# make sure this gid is even in the Graph
	sg.check_gene(gid)

	# subset t_df first, it's the easiest
	t_df = sg.t_df.loc[sg.t_df.gid == gid].copy(deep=True)
	t_df['path'] = sg.t_df.loc[sg.t_df.gid == gid].apply(
			lambda x: copy.deepcopy(x.path), axis=1)

	# subset loc_df based on all the locs that are in the paths from 
	# the already-subset t_df
	paths = t_df['path'].tolist()
	locs = [node for path in paths for node in path]
	locs = np.unique(locs)
	loc_df = sg.loc_df.loc[locs].copy(deep=True)

	# subset edge_df based on all the edges that are in the paths from 
	# the alread-subset t_df
	edges = [(v1,v2) for path in paths for v1,v2 in zip(path[:-1],path[1:])]
	edges = list(set(edges))
	edge_df = sg.edge_df.loc[edges].copy(deep=True)

	# create a new graph that's been subset
	subset_sg = Graph()
	subset_sg.loc_df = loc_df
	subset_sg.edge_df = edge_df
	subset_sg.t_df = t_df
	subset_sg.datasets = sg.datasets
	subset_sg.counts = sg.counts
	subset_sg.tpm = sg.tpm

	# renumber locs
	subset_sg.update_ids()

	# finally create the graph
	subset_sg.create_graph_from_dfs()

	return subset_sg

# convert a list of vertex ids to a list of edge ids
def vertex_to_edge_path(path):
	return [(v1,v2) for v1,v2 in zip(path[:-1],path[1:])]




def main():
	try:
	  import networkx as nx 
	  import shutil
	  import swan_vis as swan
	  nx_loc = nx.__file__
	  nx_loc = '/'.join(nx_loc.split('/')[:-1])
	  nx_loc += '/drawing/nx_pylab.py'

	  swan_loc = swan.__file__
	  swan_loc = '/'.join(swan_loc.split('/')[:-1])
	  swan_loc += '/nx_pylab.py'
	  shutil.copyfile(swan_loc, nx_loc)
	except:
  		raise Exception('Unable to patch networkx to allow for dashed edges.')	

if __name__ == '__main__': main()


from swan_vis.utils import *
from fpdf import FPDF
import matplotlib.pyplot as pyplot

# report for genes - extension of FPDF class
class Report(FPDF):

	def __init__(self,
				 prefix,
				 report_type, 
				 datasets,
				 data_type, 
				 novelty=False,
				 heatmap=False,
				 include_qvals=False):
		super().__init__(orientation='L')

		# change margins
		self.set_margins(0.5, 0.5, 0.5)

		# set report type, 'browser' or 'swan'
		if report_type != 'browser' and report_type != 'swan':
			raise Exception("Report type must be 'browser' or 'swan'.")
		else: 
			self.report_type = report_type

		# booleans of what's in the report
		self.heatmap = heatmap
		self.novelty = novelty 
		self.include_qvals = include_qvals

		# the columns that we'll include
		self.datasets = datasets
		self.report_cols = self.get_report_cols(data_type)
		self.n_dataset_cols = len(self.report_cols)
		
		# prefix for files that we'll pull from 
		self.prefix = prefix

		# color map in case we're making a heatmap
		self.cmap = plt.get_cmap('Spectral_r')

		# settings
		self.entry_height = 20

		# add extra room for the scale/colorbar if we're doing 
		# the browser/heatmap version
		if self.report_type == 'swan':
			self.header_height = 10
		if self.report_type == 'browser':
			self.header_height = 20

		# dataset width is contingent on # of datasets
		# as well as if we're including the novelty column
		if self.novelty:
			self.w_dataset = (146-25)/self.n_dataset_cols
		else:
			self.w_dataset = 146/self.n_dataset_cols

	# grab the relevant columns associated with the report type
	def get_report_cols(self, data_type):
		if not data_type:
			return self.datasets
		elif data_type == 'tpm':
			report_cols = [d+'_tpm' for d in self.datasets]
		elif data_type == 'heatmap':
			report_cols = [d+'_norm_log_tpm' for d in self.datasets]
		return report_cols

	# header - should differ based on whether it's a browser report or
	# a swan report
	def header(self):
		self.set_font('Arial', 'B', 10)
		
		# transcript ID header
		self.cell(50, self.header_height, 'Transcript ID', border=True, align='C')

		# novelty header (if needed)
		if self.novelty:
			self.cell(25, self.header_height, 'Novelty', border=True, align='C')

		# dataset ID headers
		for col in self.datasets:
			self.cell(self.w_dataset, self.header_height, col,
					  border=True, align='C')

		# in case we need to add the browser models
		browser_scale_x = self.get_x()
		browser_scale_y = self.get_y()

		# transcript model header
		self.cell(100, self.header_height, 'Transcript Model', border=True, align='C')

		# add scale if we're doing browser models
		if self.report_type == 'browser':
			self.image(self.prefix+'_browser_scale.png',
					   x=browser_scale_x, 
					   y=browser_scale_y+12,
					   w=100, h=50/7)
		self.ln()

	# footer - just add the colorbar if we're using the
	# heatmap option
	def footer(self):
		if self.heatmap:
			if not self.novelty:
				self.set_x(77.5)
			else: 
				self.set_x(90.5)
			self.image(self.prefix+'_colorbar_scale.png',
				w=90, h=135/14)

	# add a transcript model to the report
	def add_transcript(self, entry, oname):

		# entries should not be bolded
		if self.include_qvals:
			if entry.significant:
				self.set_font('Arial', 'B', 10)
			else:
				self.set_font('Arial', '', 10)
		else:
			self.set_font('Arial', '', 10)

		# tid
		self.cell(50, self.entry_height, entry['tid'], border=True, align='C')
		tid_x = self.get_x()
		tid_y = self.get_y()

		# add qvals if needed
		if self.include_qvals:
			curr_x = self.get_x()
			curr_y = self.get_y()
			if entry.significant:
				self.set_font('Arial', 'B', 6)
			else:
				self.set_font('Arial', '', 6)
			self.set_y(tid_y+12)
			text = 'qval = {:.2e}'.format(entry.qval)
			self.cell(50, 4, txt=text, border=False, align='C')
			self.set_font('Arial', '', 10)
			self.set_xy(curr_x, curr_y)

		# novelty, if necessary
		if self.novelty: 
			self.cell(25, self.entry_height, entry['novelty'],
				border=True, align='C')

		# dataset columns
		for col in self.report_cols:

			# heat map coloring
			if self.heatmap:
				color = self.cmap(entry[col])
				r = color[0]*255
				b = color[1]*255
				g = color[2]*255
				self.set_fill_color(r,b,g)
				border = False
				fill = True
				text = ''
			# TPM	
			elif '_tpm' in col:
				text = str(round(entry[col],2))
				border = True
				fill = False
			# presence/absence
			else:
				text = entry[col]
				if text == True:
					text = 'Yes'
				elif text == False:
					text = 'No'
				border = True
				fill = False 
			self.cell(self.w_dataset, self.entry_height, text,
				border=border, align='C', fill=fill)	
		x = self.get_x()
		y = self.get_y()

		# reset color to white
		self.set_fill_color(255,255,255)

		# embed transcript model
		self.cell(100, self.entry_height, '', border=True)
		self.image(oname, x=x, y=y, w=100, h=20)
		self.ln()

	# writes the pdf to file with the correct formatting
	def write_pdf(self, file):
		with open(file, 'wb') as outfile:
			outfile.write(self.output(dest='S').encode('latin-1'))



#    Copyright (C) 2004-2019 by
#    Aric Hagberg <hagberg@lanl.gov>
#    Dan Schult <dschult@colgate.edu>
#    Pieter Swart <swart@lanl.gov>
#    All rights reserved.
#    BSD license.
#
# Author: Aric Hagberg (hagberg@lanl.gov)
"""
**********
Matplotlib
**********

Draw networks with matplotlib.

See Also
--------

matplotlib:     http://matplotlib.org/

pygraphviz:     http://pygraphviz.github.io/

"""
from numbers import Number
import networkx as nx
from networkx.utils import is_string_like
from networkx.drawing.layout import shell_layout, \
    circular_layout, kamada_kawai_layout, spectral_layout, \
    spring_layout, random_layout, planar_layout

__all__ = ['draw',
           'draw_networkx',
           'draw_networkx_nodes',
           'draw_networkx_edges',
           'draw_networkx_labels',
           'draw_networkx_edge_labels',
           'draw_circular',
           'draw_kamada_kawai',
           'draw_random',
           'draw_spectral',
           'draw_spring',
           'draw_planar',
           'draw_shell']


def draw(G, pos=None, ax=None, **kwds):
    """Draw the graph G with Matplotlib.

    Draw the graph as a simple representation with no node
    labels or edge labels and using the full Matplotlib figure area
    and no axis labels by default.  See draw_networkx() for more
    full-featured drawing that allows title, axis labels etc.

    Parameters
    ----------
    G : graph
       A networkx graph

    pos : dictionary, optional
       A dictionary with nodes as keys and positions as values.
       If not specified a spring layout positioning will be computed.
       See :py:mod:`networkx.drawing.layout` for functions that
       compute node positions.

    ax : Matplotlib Axes object, optional
       Draw the graph in specified Matplotlib axes.

    kwds : optional keywords
       See networkx.draw_networkx() for a description of optional keywords.

    Examples
    --------
    >>> G = nx.dodecahedral_graph()
    >>> nx.draw(G)
    >>> nx.draw(G, pos=nx.spring_layout(G))  # use spring layout

    See Also
    --------
    draw_networkx()
    draw_networkx_nodes()
    draw_networkx_edges()
    draw_networkx_labels()
    draw_networkx_edge_labels()

    Notes
    -----
    This function has the same name as pylab.draw and pyplot.draw
    so beware when using `from networkx import *`

    since you might overwrite the pylab.draw function.

    With pyplot use

    >>> import matplotlib.pyplot as plt
    >>> import networkx as nx
    >>> G = nx.dodecahedral_graph()
    >>> nx.draw(G)  # networkx draw()
    >>> plt.draw()  # pyplot draw()

    Also see the NetworkX drawing examples at
    https://networkx.github.io/documentation/latest/auto_examples/index.html
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("Matplotlib required for draw()")
    except RuntimeError:
        print("Matplotlib unable to open display")
        raise

    if ax is None:
        cf = plt.gcf()
    else:
        cf = ax.get_figure()
    cf.set_facecolor('w')
    if ax is None:
        if cf._axstack() is None:
            ax = cf.add_axes((0, 0, 1, 1))
        else:
            ax = cf.gca()

    if 'with_labels' not in kwds:
        kwds['with_labels'] = 'labels' in kwds

    draw_networkx(G, pos=pos, ax=ax, **kwds)
    ax.set_axis_off()
    plt.draw_if_interactive()
    return


def draw_networkx(G, pos=None, arrows=True, with_labels=True, **kwds):
    """Draw the graph G using Matplotlib.

    Draw the graph with Matplotlib with options for node positions,
    labeling, titles, and many other drawing features.
    See draw() for simple drawing without labels or axes.

    Parameters
    ----------
    G : graph
       A networkx graph

    pos : dictionary, optional
       A dictionary with nodes as keys and positions as values.
       If not specified a spring layout positioning will be computed.
       See :py:mod:`networkx.drawing.layout` for functions that
       compute node positions.

    arrows : bool, optional (default=True)
       For directed graphs, if True draw arrowheads.
       Note: Arrows will be the same color as edges.

    arrowstyle : str, optional (default='-|>')
        For directed graphs, choose the style of the arrowsheads.
        See :py:class: `matplotlib.patches.ArrowStyle` for more
        options.

    arrowsize : int, optional (default=10)
       For directed graphs, choose the size of the arrow head head's length and
       width. See :py:class: `matplotlib.patches.FancyArrowPatch` for attribute
       `mutation_scale` for more info.

    with_labels :  bool, optional (default=True)
       Set to True to draw labels on the nodes.

    ax : Matplotlib Axes object, optional
       Draw the graph in the specified Matplotlib axes.

    nodelist : list, optional (default G.nodes())
       Draw only specified nodes

    edgelist : list, optional (default=G.edges())
       Draw only specified edges

    node_size : scalar or array, optional (default=300)
       Size of nodes.  If an array is specified it must be the
       same length as nodelist.

    node_color : color or array of colors (default='#1f78b4')
       Node color. Can be a single color or a sequence of colors with the same
       length as nodelist. Color can be string, or rgb (or rgba) tuple of
       floats from 0-1. If numeric values are specified they will be
       mapped to colors using the cmap and vmin,vmax parameters. See
       matplotlib.scatter for more details.

    node_shape :  string, optional (default='o')
       The shape of the node.  Specification is as matplotlib.scatter
       marker, one of 'so^>v<dph8'.

    alpha : float, optional (default=None)
       The node and edge transparency

    cmap : Matplotlib colormap, optional (default=None)
       Colormap for mapping intensities of nodes

    vmin,vmax : float, optional (default=None)
       Minimum and maximum for node colormap scaling

    linewidths : [None | scalar | sequence]
       Line width of symbol border (default =1.0)

    width : float, optional (default=1.0)
       Line width of edges

    edge_color : color or array of colors (default='k')
       Edge color. Can be a single color or a sequence of colors with the same
       length as edgelist. Color can be string, or rgb (or rgba) tuple of
       floats from 0-1. If numeric values are specified they will be
       mapped to colors using the edge_cmap and edge_vmin,edge_vmax parameters.

    edge_cmap : Matplotlib colormap, optional (default=None)
       Colormap for mapping intensities of edges

    edge_vmin,edge_vmax : floats, optional (default=None)
       Minimum and maximum for edge colormap scaling

    style : string, optional (default='solid')
       Edge line style (solid|dashed|dotted,dashdot)

    labels : dictionary, optional (default=None)
       Node labels in a dictionary keyed by node of text labels

    font_size : int, optional (default=12)
       Font size for text labels

    font_color : string, optional (default='k' black)
       Font color string

    font_weight : string, optional (default='normal')
       Font weight

    font_family : string, optional (default='sans-serif')
       Font family

    label : string, optional
        Label for graph legend

    Notes
    -----
    For directed graphs, arrows  are drawn at the head end.  Arrows can be
    turned off with keyword arrows=False.

    Examples
    --------
    >>> G = nx.dodecahedral_graph()
    >>> nx.draw(G)
    >>> nx.draw(G, pos=nx.spring_layout(G))  # use spring layout

    >>> import matplotlib.pyplot as plt
    >>> limits = plt.axis('off')  # turn of axis

    Also see the NetworkX drawing examples at
    https://networkx.github.io/documentation/latest/auto_examples/index.html

    See Also
    --------
    draw()
    draw_networkx_nodes()
    draw_networkx_edges()
    draw_networkx_labels()
    draw_networkx_edge_labels()
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("Matplotlib required for draw()")
    except RuntimeError:
        print("Matplotlib unable to open display")
        raise

    if pos is None:
        pos = nx.drawing.spring_layout(G)  # default to spring layout

    node_collection = draw_networkx_nodes(G, pos, **kwds)
    edge_collection = draw_networkx_edges(G, pos, arrows=arrows, **kwds)
    if with_labels:
        draw_networkx_labels(G, pos, **kwds)
    plt.draw_if_interactive()


def draw_networkx_nodes(G, pos,
                        nodelist=None,
                        node_size=300,
                        node_color='#1f78b4',
                        node_shape='o',
                        alpha=None,
                        cmap=None,
                        vmin=None,
                        vmax=None,
                        ax=None,
                        linewidths=None,
                        edgecolors=None,
                        label=None,
                        **kwds):
    """Draw the nodes of the graph G.

    This draws only the nodes of the graph G.

    Parameters
    ----------
    G : graph
       A networkx graph

    pos : dictionary
       A dictionary with nodes as keys and positions as values.
       Positions should be sequences of length 2.

    ax : Matplotlib Axes object, optional
       Draw the graph in the specified Matplotlib axes.

    nodelist : list, optional
       Draw only specified nodes (default G.nodes())

    node_size : scalar or array
       Size of nodes (default=300).  If an array is specified it must be the
       same length as nodelist.

    node_color : color or array of colors (default='#1f78b4')
       Node color. Can be a single color or a sequence of colors with the same
       length as nodelist. Color can be string, or rgb (or rgba) tuple of
       floats from 0-1. If numeric values are specified they will be
       mapped to colors using the cmap and vmin,vmax parameters. See
       matplotlib.scatter for more details.

    node_shape :  string
       The shape of the node.  Specification is as matplotlib.scatter
       marker, one of 'so^>v<dph8' (default='o').

    alpha : float or array of floats
       The node transparency.  This can be a single alpha value (default=None),
       in which case it will be applied to all the nodes of color. Otherwise,
       if it is an array, the elements of alpha will be applied to the colors
       in order (cycling through alpha multiple times if necessary).

    cmap : Matplotlib colormap
       Colormap for mapping intensities of nodes (default=None)

    vmin,vmax : floats
       Minimum and maximum for node colormap scaling (default=None)

    linewidths : [None | scalar | sequence]
       Line width of symbol border (default =1.0)

    edgecolors : [None | scalar | sequence]
       Colors of node borders (default = node_color)

    label : [None| string]
       Label for legend

    min_source_margin : int, optional (default=0)
       The minimum margin (gap) at the begining of the edge at the source.

    min_target_margin : int, optional (default=0)
       The minimum margin (gap) at the end of the edge at the target.

    Returns
    -------
    matplotlib.collections.PathCollection
        `PathCollection` of the nodes.

    Examples
    --------
    >>> G = nx.dodecahedral_graph()
    >>> nodes = nx.draw_networkx_nodes(G, pos=nx.spring_layout(G))

    Also see the NetworkX drawing examples at
    https://networkx.github.io/documentation/latest/auto_examples/index.html

    See Also
    --------
    draw()
    draw_networkx()
    draw_networkx_edges()
    draw_networkx_labels()
    draw_networkx_edge_labels()
    """
    from collections.abc import Iterable
    try:
        import matplotlib.pyplot as plt
        import numpy as np
    except ImportError:
        raise ImportError("Matplotlib required for draw()")
    except RuntimeError:
        print("Matplotlib unable to open display")
        raise

    if ax is None:
        ax = plt.gca()

    if nodelist is None:
        nodelist = list(G)

    if len(nodelist) == 0:  # empty nodelist, no drawing
        return

    try:
        xy = np.asarray([pos[v] for v in nodelist])
    except KeyError as e:
        raise nx.NetworkXError('Node %s has no position.' % e)
    except ValueError:
        raise nx.NetworkXError('Bad value in node positions.')

    if isinstance(alpha, Iterable):
        node_color = apply_alpha(node_color, alpha, nodelist, cmap, vmin, vmax)
        alpha = None

    node_collection = ax.scatter(xy[:, 0], xy[:, 1],
                                 s=node_size,
                                 c=node_color,
                                 marker=node_shape,
                                 cmap=cmap,
                                 vmin=vmin,
                                 vmax=vmax,
                                 alpha=alpha,
                                 linewidths=linewidths,
                                 edgecolors=edgecolors,
                                 label=label)
    ax.tick_params(
        axis='both',
        which='both',
        bottom=False,
        left=False,
        labelbottom=False,
        labelleft=False)

    node_collection.set_zorder(2)
    return node_collection


def draw_networkx_edges(G, pos,
                        edgelist=None,
                        width=1.0,
                        edge_color='k',
                        style='solid',
                        alpha=None,
                        arrowstyle='-|>',
                        arrowsize=10,
                        edge_cmap=None,
                        edge_vmin=None,
                        edge_vmax=None,
                        ax=None,
                        arrows=True,
                        label=None,
                        node_size=300,
                        nodelist=None,
                        node_shape="o",
                        connectionstyle=None,
                        min_source_margin=0,
                        min_target_margin=0,
                        **kwds):
    """Draw the edges of the graph G.

    This draws only the edges of the graph G.

    Parameters
    ----------
    G : graph
       A networkx graph

    pos : dictionary
       A dictionary with nodes as keys and positions as values.
       Positions should be sequences of length 2.

    edgelist : collection of edge tuples
       Draw only specified edges(default=G.edges())

    width : float, or array of floats
       Line width of edges (default=1.0)

    edge_color : color or array of colors (default='k')
       Edge color. Can be a single color or a sequence of colors with the same
       length as edgelist. Color can be string, or rgb (or rgba) tuple of
       floats from 0-1. If numeric values are specified they will be
       mapped to colors using the edge_cmap and edge_vmin,edge_vmax parameters.

    style : string
       Edge line style (default='solid') (solid|dashed|dotted,dashdot)

    alpha : float
       The edge transparency (default=None)

    edge_ cmap : Matplotlib colormap
       Colormap for mapping intensities of edges (default=None)

    edge_vmin,edge_vmax : floats
       Minimum and maximum for edge colormap scaling (default=None)

    ax : Matplotlib Axes object, optional
       Draw the graph in the specified Matplotlib axes.

    arrows : bool, optional (default=True)
       For directed graphs, if True draw arrowheads.
       Note: Arrows will be the same color as edges.

    arrowstyle : str, optional (default='-|>')
       For directed graphs, choose the style of the arrow heads.
       See :py:class: `matplotlib.patches.ArrowStyle` for more
       options.

    arrowsize : int, optional (default=10)
       For directed graphs, choose the size of the arrow head head's length and
       width. See :py:class: `matplotlib.patches.FancyArrowPatch` for attribute
       `mutation_scale` for more info.

    connectionstyle : str, optional (default=None)
       Pass the connectionstyle parameter to create curved arc of rounding
       radius rad. For example, connectionstyle='arc3,rad=0.2'.
       See :py:class: `matplotlib.patches.ConnectionStyle` and
       :py:class: `matplotlib.patches.FancyArrowPatch` for more info.

    label : [None| string]
       Label for legend

    min_source_margin : int, optional (default=0)
       The minimum margin (gap) at the begining of the edge at the source.

    min_target_margin : int, optional (default=0)
       The minimum margin (gap) at the end of the edge at the target.

    Returns
    -------
    matplotlib.collection.LineCollection
        `LineCollection` of the edges

    list of matplotlib.patches.FancyArrowPatch
        `FancyArrowPatch` instances of the directed edges

    Depending whether the drawing includes arrows or not.

    Notes
    -----
    For directed graphs, arrows are drawn at the head end.  Arrows can be
    turned off with keyword arrows=False. Be sure to include `node_size` as a
    keyword argument; arrows are drawn considering the size of nodes.

    Examples
    --------
    >>> G = nx.dodecahedral_graph()
    >>> edges = nx.draw_networkx_edges(G, pos=nx.spring_layout(G))

    >>> G = nx.DiGraph()
    >>> G.add_edges_from([(1, 2), (1, 3), (2, 3)])
    >>> arcs = nx.draw_networkx_edges(G, pos=nx.spring_layout(G))
    >>> alphas = [0.3, 0.4, 0.5]
    >>> for i, arc in enumerate(arcs):  # change alpha values of arcs
    ...     arc.set_alpha(alphas[i])

    Also see the NetworkX drawing examples at
    https://networkx.github.io/documentation/latest/auto_examples/index.html

    See Also
    --------
    draw()
    draw_networkx()
    draw_networkx_nodes()
    draw_networkx_labels()
    draw_networkx_edge_labels()
    """
    try:
        import matplotlib
        import matplotlib.pyplot as plt
        from matplotlib.colors import colorConverter, Colormap, Normalize
        from matplotlib.collections import LineCollection
        from matplotlib.patches import FancyArrowPatch
        import numpy as np
    except ImportError:
        raise ImportError("Matplotlib required for draw()")
    except RuntimeError:
        print("Matplotlib unable to open display")
        raise

    if ax is None:
        ax = plt.gca()

    if edgelist is None:
        edgelist = list(G.edges())

    if not edgelist or len(edgelist) == 0:  # no edges!
        return None

    if nodelist is None:
        nodelist = list(G.nodes())

    # FancyArrowPatch handles color=None different from LineCollection
    if edge_color is None:
        edge_color = 'k'

    # set edge positions
    edge_pos = np.asarray([(pos[e[0]], pos[e[1]]) for e in edgelist])

    # Check if edge_color is an array of floats and map to edge_cmap.
    # This is the only case handled differently from matplotlib
    if np.iterable(edge_color) and (len(edge_color) == len(edge_pos)) \
            and np.alltrue([isinstance(c, Number) for c in edge_color]):
        if edge_cmap is not None:
            assert(isinstance(edge_cmap, Colormap))
        else:
            edge_cmap = plt.get_cmap()
        if edge_vmin is None:
            edge_vmin = min(edge_color)
        if edge_vmax is None:
            edge_vmax = max(edge_color)
        color_normal = Normalize(vmin=edge_vmin, vmax=edge_vmax)
        edge_color = [edge_cmap(color_normal(e)) for e in edge_color]

    if (not G.is_directed() or not arrows):
        edge_collection = LineCollection(edge_pos,
                                         colors=edge_color,
                                         linewidths=width,
                                         antialiaseds=(1,),
                                         linestyle=style,
                                         transOffset=ax.transData,
                                         alpha=alpha
                                         )

        edge_collection.set_zorder(1)  # edges go behind nodes
        edge_collection.set_label(label)
        ax.add_collection(edge_collection)

        return edge_collection

    arrow_collection = None

    if G.is_directed() and arrows:
        # Note: Waiting for someone to implement arrow to intersection with
        # marker.  Meanwhile, this works well for polygons with more than 4
        # sides and circle.

        def to_marker_edge(marker_size, marker):
            if marker in "s^>v<d":  # `large` markers need extra space
                return np.sqrt(2 * marker_size) / 2
            else:
                return np.sqrt(marker_size) / 2

        # Draw arrows with `matplotlib.patches.FancyarrowPatch`
        arrow_collection = []
        mutation_scale = arrowsize  # scale factor of arrow head

        # FancyArrowPatch doesn't handle color strings
        arrow_colors = colorConverter.to_rgba_array(edge_color, alpha)
        for i, (src, dst) in enumerate(edge_pos):
            x1, y1 = src
            x2, y2 = dst
            shrink_source = 0  # space from source to tail
            shrink_target = 0  # space from  head to target
            if np.iterable(node_size):  # many node sizes
                src_node, dst_node = edgelist[i][:2]
                index_node = nodelist.index(dst_node)
                marker_size = node_size[index_node]
                shrink_target = to_marker_edge(marker_size, node_shape)
            else:
                shrink_target = to_marker_edge(node_size, node_shape)

            if shrink_source < min_source_margin:
                shrink_source = min_source_margin

            if shrink_target < min_target_margin:
                shrink_target = min_target_margin

            if len(arrow_colors) == len(edge_pos):
                arrow_color = arrow_colors[i]
            elif len(arrow_colors) == 1:
                arrow_color = arrow_colors[0]
            else:  # Cycle through colors
                arrow_color = arrow_colors[i % len(arrow_colors)]

            if np.iterable(width):
                if len(width) == len(edge_pos):
                    line_width = width[i]
                else:
                    line_width = width[i % len(width)]
            else:
                line_width = width

            arrow = FancyArrowPatch((x1, y1), (x2, y2),
                                    arrowstyle=arrowstyle,
                                    shrinkA=shrink_source,
                                    shrinkB=shrink_target,
                                    mutation_scale=mutation_scale,
                                    color=arrow_color,
                                    linewidth=line_width,
                                    connectionstyle=connectionstyle,
                                    linestyle=style,
                                    zorder=1)  # arrows go behind nodes

            # There seems to be a bug in matplotlib to make collections of
            # FancyArrowPatch instances. Until fixed, the patches are added
            # individually to the axes instance.
            arrow_collection.append(arrow)
            ax.add_patch(arrow)

    # update view
    minx = np.amin(np.ravel(edge_pos[:, :, 0]))
    maxx = np.amax(np.ravel(edge_pos[:, :, 0]))
    miny = np.amin(np.ravel(edge_pos[:, :, 1]))
    maxy = np.amax(np.ravel(edge_pos[:, :, 1]))

    w = maxx - minx
    h = maxy - miny
    padx,  pady = 0.05 * w, 0.05 * h
    corners = (minx - padx, miny - pady), (maxx + padx, maxy + pady)
    ax.update_datalim(corners)
    ax.autoscale_view()

    ax.tick_params(
        axis='both',
        which='both',
        bottom=False,
        left=False,
        labelbottom=False,
        labelleft=False)

    return arrow_collection


def draw_networkx_labels(G, pos,
                         labels=None,
                         font_size=12,
                         font_color='k',
                         font_family='sans-serif',
                         font_weight='normal',
                         alpha=None,
                         bbox=None,
                         ax=None,
                         **kwds):
    """Draw node labels on the graph G.

    Parameters
    ----------
    G : graph
       A networkx graph

    pos : dictionary
       A dictionary with nodes as keys and positions as values.
       Positions should be sequences of length 2.

    labels : dictionary, optional (default=None)
       Node labels in a dictionary keyed by node of text labels
       Node-keys in labels should appear as keys in `pos`.
       If needed use: `{n:lab for n,lab in labels.items() if n in pos}`

    font_size : int
       Font size for text labels (default=12)

    font_color : string
       Font color string (default='k' black)

    font_family : string
       Font family (default='sans-serif')

    font_weight : string
       Font weight (default='normal')

    alpha : float or None
       The text transparency (default=None)

    ax : Matplotlib Axes object, optional
       Draw the graph in the specified Matplotlib axes.

    Returns
    -------
    dict
        `dict` of labels keyed on the nodes

    Examples
    --------
    >>> G = nx.dodecahedral_graph()
    >>> labels = nx.draw_networkx_labels(G, pos=nx.spring_layout(G))

    Also see the NetworkX drawing examples at
    https://networkx.github.io/documentation/latest/auto_examples/index.html

    See Also
    --------
    draw()
    draw_networkx()
    draw_networkx_nodes()
    draw_networkx_edges()
    draw_networkx_edge_labels()
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError("Matplotlib required for draw()")
    except RuntimeError:
        print("Matplotlib unable to open display")
        raise

    if ax is None:
        ax = plt.gca()

    if labels is None:
        labels = dict((n, n) for n in G.nodes())

    # set optional alignment
    horizontalalignment = kwds.get('horizontalalignment', 'center')
    verticalalignment = kwds.get('verticalalignment', 'center')

    text_items = {}  # there is no text collection so we'll fake one
    for n, label in labels.items():
        (x, y) = pos[n]
        if not is_string_like(label):
            label = str(label)  # this makes "1" and 1 labeled the same
        t = ax.text(x, y,
                    label,
                    size=font_size,
                    color=font_color,
                    family=font_family,
                    weight=font_weight,
                    alpha=alpha,
                    horizontalalignment=horizontalalignment,
                    verticalalignment=verticalalignment,
                    transform=ax.transData,
                    bbox=bbox,
                    clip_on=True,
                    )
        text_items[n] = t

    ax.tick_params(
        axis='both',
        which='both',
        bottom=False,
        left=False,
        labelbottom=False,
        labelleft=False)

    return text_items


def draw_networkx_edge_labels(G, pos,
                              edge_labels=None,
                              label_pos=0.5,
                              font_size=10,
                              font_color='k',
                              font_family='sans-serif',
                              font_weight='normal',
                              alpha=None,
                              bbox=None,
                              ax=None,
                              rotate=True,
                              **kwds):
    """Draw edge labels.

    Parameters
    ----------
    G : graph
       A networkx graph

    pos : dictionary
       A dictionary with nodes as keys and positions as values.
       Positions should be sequences of length 2.

    ax : Matplotlib Axes object, optional
       Draw the graph in the specified Matplotlib axes.

    alpha : float or None
       The text transparency (default=None)

    edge_labels : dictionary
       Edge labels in a dictionary keyed by edge two-tuple of text
       labels (default=None). Only labels for the keys in the dictionary
       are drawn.

    label_pos : float
       Position of edge label along edge (0=head, 0.5=center, 1=tail)

    font_size : int
       Font size for text labels (default=12)

    font_color : string
       Font color string (default='k' black)

    font_weight : string
       Font weight (default='normal')

    font_family : string
       Font family (default='sans-serif')

    bbox : Matplotlib bbox
       Specify text box shape and colors.

    clip_on : bool
       Turn on clipping at axis boundaries (default=True)

    Returns
    -------
    dict
        `dict` of labels keyed on the edges

    Examples
    --------
    >>> G = nx.dodecahedral_graph()
    >>> edge_labels = nx.draw_networkx_edge_labels(G, pos=nx.spring_layout(G))

    Also see the NetworkX drawing examples at
    https://networkx.github.io/documentation/latest/auto_examples/index.html

    See Also
    --------
    draw()
    draw_networkx()
    draw_networkx_nodes()
    draw_networkx_edges()
    draw_networkx_labels()
    """
    try:
        import matplotlib.pyplot as plt
        import numpy as np
    except ImportError:
        raise ImportError("Matplotlib required for draw()")
    except RuntimeError:
        print("Matplotlib unable to open display")
        raise

    if ax is None:
        ax = plt.gca()
    if edge_labels is None:
        labels = {(u, v): d for u, v, d in G.edges(data=True)}
    else:
        labels = edge_labels
    text_items = {}
    for (n1, n2), label in labels.items():
        (x1, y1) = pos[n1]
        (x2, y2) = pos[n2]
        (x, y) = (x1 * label_pos + x2 * (1.0 - label_pos),
                  y1 * label_pos + y2 * (1.0 - label_pos))

        if rotate:
            # in degrees
            angle = np.arctan2(y2 - y1, x2 - x1) / (2.0 * np.pi) * 360
            # make label orientation "right-side-up"
            if angle > 90:
                angle -= 180
            if angle < - 90:
                angle += 180
            # transform data coordinate angle to screen coordinate angle
            xy = np.array((x, y))
            trans_angle = ax.transData.transform_angles(np.array((angle,)),
                                                        xy.reshape((1, 2)))[0]
        else:
            trans_angle = 0.0
        # use default box of white with white border
        if bbox is None:
            bbox = dict(boxstyle='round',
                        ec=(1.0, 1.0, 1.0),
                        fc=(1.0, 1.0, 1.0),
                        )
        if not is_string_like(label):
            label = str(label)  # this makes "1" and 1 labeled the same

        # set optional alignment
        horizontalalignment = kwds.get('horizontalalignment', 'center')
        verticalalignment = kwds.get('verticalalignment', 'center')

        t = ax.text(x, y,
                    label,
                    size=font_size,
                    color=font_color,
                    family=font_family,
                    weight=font_weight,
                    alpha=alpha,
                    horizontalalignment=horizontalalignment,
                    verticalalignment=verticalalignment,
                    rotation=trans_angle,
                    transform=ax.transData,
                    bbox=bbox,
                    zorder=1,
                    clip_on=True,
                    )
        text_items[(n1, n2)] = t

    ax.tick_params(
        axis='both',
        which='both',
        bottom=False,
        left=False,
        labelbottom=False,
        labelleft=False)

    return text_items


def draw_circular(G, **kwargs):
    """Draw the graph G with a circular layout.

    Parameters
    ----------
    G : graph
       A networkx graph

    kwargs : optional keywords
       See networkx.draw_networkx() for a description of optional keywords,
       with the exception of the pos parameter which is not used by this
       function.
    """
    draw(G, circular_layout(G), **kwargs)


def draw_kamada_kawai(G, **kwargs):
    """Draw the graph G with a Kamada-Kawai force-directed layout.

    Parameters
    ----------
    G : graph
       A networkx graph

    kwargs : optional keywords
       See networkx.draw_networkx() for a description of optional keywords,
       with the exception of the pos parameter which is not used by this
       function.
    """
    draw(G, kamada_kawai_layout(G), **kwargs)


def draw_random(G, **kwargs):
    """Draw the graph G with a random layout.

    Parameters
    ----------
    G : graph
       A networkx graph

    kwargs : optional keywords
       See networkx.draw_networkx() for a description of optional keywords,
       with the exception of the pos parameter which is not used by this
       function.
    """
    draw(G, random_layout(G), **kwargs)


def draw_spectral(G, **kwargs):
    """Draw the graph G with a spectral 2D layout.

    Using the unnormalized Laplacion, the layout shows possible clusters of
    nodes which are an approximation of the ratio cut. The positions are the
    entries of the second and third eigenvectors corresponding to the
    ascending eigenvalues starting from the second one.

    Parameters
    ----------
    G : graph
       A networkx graph

    kwargs : optional keywords
       See networkx.draw_networkx() for a description of optional keywords,
       with the exception of the pos parameter which is not used by this
       function.
    """
    draw(G, spectral_layout(G), **kwargs)


def draw_spring(G, **kwargs):
    """Draw the graph G with a spring layout.

    Parameters
    ----------
    G : graph
       A networkx graph

    kwargs : optional keywords
       See networkx.draw_networkx() for a description of optional keywords,
       with the exception of the pos parameter which is not used by this
       function.
    """
    draw(G, spring_layout(G), **kwargs)


def draw_shell(G, **kwargs):
    """Draw networkx graph with shell layout.

    Parameters
    ----------
    G : graph
       A networkx graph

    kwargs : optional keywords
       See networkx.draw_networkx() for a description of optional keywords,
       with the exception of the pos parameter which is not used by this
       function.
    """
    nlist = kwargs.get('nlist', None)
    if nlist is not None:
        del(kwargs['nlist'])
    draw(G, shell_layout(G, nlist=nlist), **kwargs)


def draw_planar(G, **kwargs):
    """Draw a planar networkx graph with planar layout.

    Parameters
    ----------
    G : graph
       A planar networkx graph

    kwargs : optional keywords
       See networkx.draw_networkx() for a description of optional keywords,
       with the exception of the pos parameter which is not used by this
       function.
    """
    draw(G, planar_layout(G), **kwargs)


def apply_alpha(colors, alpha, elem_list, cmap=None, vmin=None, vmax=None):
    """Apply an alpha (or list of alphas) to the colors provided.

    Parameters
    ----------

    colors : color string, or array of floats
       Color of element. Can be a single color format string (default='r'),
       or a  sequence of colors with the same length as nodelist.
       If numeric values are specified they will be mapped to
       colors using the cmap and vmin,vmax parameters.  See
       matplotlib.scatter for more details.

    alpha : float or array of floats
       Alpha values for elements. This can be a single alpha value, in
       which case it will be applied to all the elements of color. Otherwise,
       if it is an array, the elements of alpha will be applied to the colors
       in order (cycling through alpha multiple times if necessary).

    elem_list : array of networkx objects
       The list of elements which are being colored. These could be nodes,
       edges or labels.

    cmap : matplotlib colormap
       Color map for use if colors is a list of floats corresponding to points
       on a color mapping.

    vmin, vmax : float
       Minimum and maximum values for normalizing colors if a color mapping is
       used.

    Returns
    -------

    rgba_colors : numpy ndarray
        Array containing RGBA format values for each of the node colours.

    """
    from itertools import islice, cycle

    try:
        import numpy as np
        from matplotlib.colors import colorConverter
        import matplotlib.cm as cm
    except ImportError:
        raise ImportError("Matplotlib required for draw()")

    # If we have been provided with a list of numbers as long as elem_list,
    # apply the color mapping.
    if len(colors) == len(elem_list) and isinstance(colors[0], Number):
        mapper = cm.ScalarMappable(cmap=cmap)
        mapper.set_clim(vmin, vmax)
        rgba_colors = mapper.to_rgba(colors)
    # Otherwise, convert colors to matplotlib's RGB using the colorConverter
    # object.  These are converted to numpy ndarrays to be consistent with the
    # to_rgba method of ScalarMappable.
    else:
        try:
            rgba_colors = np.array([colorConverter.to_rgba(colors)])
        except ValueError:
            rgba_colors = np.array([colorConverter.to_rgba(color)
                                    for color in colors])
    # Set the final column of the rgba_colors to have the relevant alpha values
    try:
        # If alpha is longer than the number of colors, resize to the number of
        # elements.  Also, if rgba_colors.size (the number of elements of
        # rgba_colors) is the same as the number of elements, resize the array,
        # to avoid it being interpreted as a colormap by scatter()
        if len(alpha) > len(rgba_colors) or rgba_colors.size == len(elem_list):
            rgba_colors = np.resize(rgba_colors, (len(elem_list), 4))
            rgba_colors[1:, 0] = rgba_colors[0, 0]
            rgba_colors[1:, 1] = rgba_colors[0, 1]
            rgba_colors[1:, 2] = rgba_colors[0, 2]
        rgba_colors[:,  3] = list(islice(cycle(alpha), len(rgba_colors)))
    except TypeError:
        rgba_colors[:, -1] = alpha
    return rgba_colors


# fixture for pytest
def setup_module(module):
    import pytest
    mpl = pytest.importorskip('matplotlib')
    mpl.use('PS', warn=False)
    plt = pytest.importorskip('matplotlib.pyplot')



import pytest
import sys
import numpy as np
import swan_vis as swan
import networkx as nx
import math
import pandas as pd

class TestSpliceGraph(object):
	
	###########################################################################
	######################### Merging tests ###################################
	###########################################################################


	# TODO add_dataset test
	# def test_add_dataset(self):
	# 	a = get_dummy_sg()

	# TODO add_annotation test

	# TODO add_dataset (with abundance) test

	# test for merge_dfs, the following is a summative assesment of 
	# many methods, individual tests for which are also present in this file
	def test_merge_dfs(self):

		a,b = get_dummy_merge_sgs()
		print(a.loc_df)
		print(b.loc_df)

		a.merge_dfs(b, 'b')

		# loc_df
		print('loc_df')
		test = a.loc_df.apply(
			lambda x: (x.vertex_id, x.a, x.b), axis=1)
		control = [(0,True,True),(1,True,True),(2,True,False),
				   (3,True,False),(4,False,True),(5,False,True),
				   (6,False,True)]
		check_pairs(control, test)

		# edge_df
		print('edge_df')
		test = a.edge_df.apply(
			lambda x: (x.edge_id, x.a, x.b), axis=1)
		control = [((0,1),True,True),((1,2),True, False),((0,2),True,False),
			       ((2,3),True,False),((0,4),False,True),((1,5),False,True),
				   ((4,5),False,True)]
		check_pairs(control, test)

		# t_df
		print('t_df')
		test = a.t_df.apply(
				lambda x: (x.tid, x.a, x.b), axis=1)
		control = [(0,True,True),(1,True,False),(2,False,True),
				   (3,True,False),(4,False,True)]
		check_pairs(control, test)

	# test merge loc_dfs
	def test_merge_loc_dfs(self):
		a,b = get_dummy_merge_sgs()

		a.merge_loc_dfs(b, 'b')

		test = a.loc_df.apply(
			lambda x: (x.vertex_id_a, x.vertex_id_b), axis=1)
		test = [nan_to_neg(pair) for pair in test]
		control = [(0,1),(1,2),(2,np.nan),(3,np.nan),(np.nan,3),
				   (np.nan,4),(np.nan,5)]
		control = [nan_to_neg(pair) for pair in control]
		check_pairs(control, test)

	def test_get_merged_id_map(self):
		a,b = get_dummy_merge_sgs()
		a.merge_loc_dfs(b, 'b')
		id_map = a.get_merged_id_map()

		test = list(id_map.items())
		control = [(1,0),(2,1),(3,4),(4,5),(5,6)]
		check_pairs(control, test)


	############################################################################
	####################### Creation utility tests #############################
	############################################################################
	
	# TODO test create_dfs_gtf (will need dummy input file)

	# TODO test create_dfs_db (will need dummy input file)

	def test_get_loc_types(self):
		a = get_dummy_sg('no_locs')
		a.get_loc_types()

		control_tss = [0,1]
		control_tes = [1,2]
		control_internal = [1]

		test_tss = a.loc_df[a.loc_df.apply(
			lambda x: True if x.TSS else False, axis=1)].vertex_id.tolist()
		test_tes = a.loc_df[a.loc_df.apply(
			lambda x: True if x.TES else False, axis=1)].vertex_id.tolist()
		test_internal = a.loc_df[a.loc_df.apply(
			lambda x: True if x.internal else False, axis=1)].vertex_id.tolist()

		print('tss')
		check_pairs(control_tss, test_tss)
		print('tes')
		check_pairs(control_tes, test_tes)
		print('internal')
		check_pairs(control_internal, test_internal)

	############################################################################
	###################### Ordering transcripts tests ##########################
	############################################################################

	def test_order_transcripts(self):
		a = get_dummy_sg()
		a.t_df['a_tpm'] =[0,0,12]
		a.t_df['b_tpm'] = [1,0,14]
		a.t_df['a_tpm']
		a.tpm = ['a_tpm', 'b_tpm']

		# order by expression level
		a.order_transcripts(order='expression')
		print(a.t_df.head())
		assert a.t_df.tid.tolist() == [0,2,1]

		# order by transcript id
		a.order_transcripts()
		print(a.t_df.head())
		assert a.t_df.tid.tolist() == [0,1,2]

		# order by coordinate of tss
		a.order_transcripts(order='tss')
		print(a.t_df.head())
		assert a.t_df.tid.tolist() == [0,2,1]

		# order by coordinate of tes
		a.order_transcripts()
		a.order_transcripts(order='tes')

		print()
		print(a.loc_df)
		print(a.edge_df)
		print(a.t_df)
		assert a.t_df.tid.tolist() == [1,2,0]

	############################################################################
	####################### Adding counts/tpms test ############################
	############################################################################

	def test_add_abundance(self):
		a = get_dummy_sg(special='dataset')
		a.add_abundance('input_files/test_abundance.tsv',
			['count_1a', 'count_2a'], 'dataset_a')

		assert 'dataset_a_counts' in a.t_df.columns
		assert 'dataset_a_tpm' in a.t_df.columns
		assert a.get_count_cols() == ['dataset_a_counts']
		assert a.get_tpm_cols() == ['dataset_a_tpm']

def nan_to_neg(pair):
	new_tuple = []
	for i in pair:
		if math.isnan(i):
			new_tuple.append(-1)
		else: 
			new_tuple.append(i)
	return tuple(new_tuple)


def check_pairs(control, test):
	print('control')
	print(control)
	print('test')
	print(test)
	for t in test:
		assert t in control
	assert len(test) == len(control)

def get_dummy_sg(special=None):
	a = swan.SwanGraph()

	loc_df = pd.DataFrame({'chrom': [1,1,1],
		'coord': [1,3,2],
		'strand': ['+', '+', '+'],
		'vertex_id': [0,1,2]})
	loc_df = swan.create_dupe_index(loc_df, 'vertex_id')
	loc_df = swan.set_dupe_index(loc_df, 'vertex_id')

	edge_df = pd.DataFrame({'edge_id': [(0,2),(0,1),(1,2)],
			  'v1': [0,0,1],
			  'v2': [2,1,2],
			  'edge_type': ['exon', 'exon', 'intron'], 
			  'strand': ['+','+','+']})
	edge_df = swan.create_dupe_index(edge_df, 'edge_id')
	edge_df = swan.set_dupe_index(edge_df, 'edge_id')

	t_df = pd.DataFrame({'tid':[2,1,0],
			  'gid':[0,0,0],
			  'gname':['0','0','0'],
			  'path':[[0,1,2],[1,2],[0,1]],
			  'counts_a': [0,0,12],
			  'counts_b': [1,0,14]})
	t_df = swan.create_dupe_index(t_df, 'tid')
	t_df = swan.set_dupe_index(t_df, 'tid')

	if special == 'intron':
		edge_df.loc[(0,1), 'edge_type'] = 'intron'


	a.loc_df = loc_df
	a.edge_df = edge_df
	a.t_df = t_df 

	if special == 'no_locs':	
		pass
	else:
		a.get_loc_types()

	if special == 'dataset':
		a.datasets = ['dataset_a']
		a.loc_df['dataset_a'] = True
		a.edge_df['dataset_a'] = True
		a.t_df['dataset_a'] = True

	return a

def get_dummy_merge_sgs():
		a = swan.SwanGraph()
		b = swan.SwanGraph()

		a.loc_df = pd.DataFrame({'chrom': [1,1,1,2],
								 'coord': [1,2,3,1],
								 'strand': ['+', '+', '+', '+'],
								 'vertex_id': [0,1,2,3]})
		b.loc_df = pd.DataFrame({'chrom': [1,1,1,2,3],
								 'coord': [1,2,4,1,1],
								 'strand': ['+', '+', '+', '-', '+'],
								 'vertex_id': [1,2,3,4,5]})
		a.loc_df = swan.create_dupe_index(a.loc_df, 'vertex_id')
		b.loc_df = swan.create_dupe_index(b.loc_df, 'vertex_id')
		a.loc_df = swan.set_dupe_index(a.loc_df, 'vertex_id')
		b.loc_df = swan.set_dupe_index(b.loc_df, 'vertex_id')

		a.edge_df = pd.DataFrame({'edge_id': [(0,1),(1,2),(0,2),(2,3)],
								  'v1': [0,1,0,2],
								  'v2': [1,2,2,3],
								  'edge_type': ['exon', 'intron', 'exon', 'exon'], 
								  'strand': ['+','+','+','+']})
		b.edge_df = pd.DataFrame({'edge_id': [(1,2),(1,3),(2,4),(3,4)],
								  'v1': [1,1,2,3],
								  'v2': [2,3,4,4],
								  'strand': ['+','+','+','+'],
								  'edge_type': ['exon', 'exon', 'intron', 'intron']})
		a.edge_df = swan.create_dupe_index(a.edge_df, 'edge_id')
		b.edge_df = swan.create_dupe_index(b.edge_df, 'edge_id')
		a.edge_df = swan.set_dupe_index(a.edge_df, 'edge_id')
		b.edge_df = swan.set_dupe_index(b.edge_df, 'edge_id')

		a.t_df = pd.DataFrame({'tid': [0,1,3],
								'gid': [0,0,0],
								'gname': ['0','0','0'],
								'path': [[0,1],[0,1,2,3],[0,2]]})
		b.t_df = pd.DataFrame({'tid': [0,2,4],
								'gid': [0,0,0],
								'gname': ['0','0','0'],
								'path': [[1,2],[1,2,4],[1,2,3]]})
		a.t_df = swan.create_dupe_index(a.t_df, 'tid')
		b.t_df = swan.create_dupe_index(b.t_df, 'tid')
		a.t_df = swan.set_dupe_index(a.t_df, 'tid')
		b.t_df = swan.set_dupe_index(b.t_df, 'tid')

		# add 'dataset a' to a
		a.datasets = ['a']
		a.loc_df['a'] = True
		a.edge_df['a'] = True
		a.t_df['a'] = True

		a.get_loc_types()
		b.get_loc_types()


		return a, b







import pytest
import sys
import numpy as np
import swan_vis as swan
import networkx as nx
import math
import pandas as pd

# tests the functions in superclass Graph.py
class TestGraph(object):

	##########################################################################
	################# Related to checking contents of Graph ##################
	##########################################################################

	# test for check_datasets
	def test_check_datasets(self):

		g = swan.SwanGraph()
		g.datasets = ['a','b']

		# 1: test for a dataset that's not there
		with pytest.raises(Exception) as excinfo:
			g.check_datasets(['c'])
		assert "c not present in graph" in str(excinfo.value)

		# 2: test for a dataset that is there, input in string form
		result = g.check_datasets('a')
		assert result == None

		# 3: test for multiple datasets that are in the graph
		result = g.check_datasets(['a', 'b'])
		assert result == None

		# 4: test for a dataset that's not there in a list 
		# of datasets that are there
		with pytest.raises(Exception) as excinfo:
			g.check_datasets(['a', 'b', 'c'])
		assert 'c not present in graph' in str(excinfo.value)

	# test for check_abundances
	def test_check_abundances(self):
		g = swan.SwanGraph()
		g.datasets = ['a', 'b', 'c']
		g.counts = ['a_counts', 'b_counts']

		# 1: test for abundance that isn't there
		# also makes sure it works with character input (ie not list)
		with pytest.raises(Exception) as excinfo:
			g.check_abundances('d')
		assert 'Abundance for dataset d' in str(excinfo.value)

		# 2: test for a abundance from a dataset that is there but
		# does not have abundance info
		# also makes sure it works with len(list) = 1 info
		with pytest.raises(Exception) as excinfo:
			g.check_abundances(['c'])			
		assert 'Abundance for dataset c' in str(excinfo.value)

		# 3: test for multiple datasets that are in the graph
		result = g.check_abundances(['a', 'b'])
		assert result == None

		# 4: test for a dataset that's not there in a list 
		# of datasets that are there
		with pytest.raises(Exception) as excinfo:
			g.check_abundances(['a', 'b', 'd'])
		assert 'Abundance for dataset d' in str(excinfo.value)

	##########################################################################
	####################### Related to creating Graph ########################
	##########################################################################

	# test update_ids
	def test_update_ids(self):
		a = get_dummy_sg()
		a.update_ids()

		# loc_df
		print('loc_df')
		test = a.loc_df.apply(
			lambda x: (x.vertex_id, x.coord), axis=1)
		control = [(0,1),(1,2),(2,3)]
		check_pairs(control, test)

		# edge_df
		print('edge_df')
		test = a.edge_df.edge_id.tolist()
		control = [(0,1),(0,2),(2,1)]
		check_pairs(control, test)

		# t_df
		print('t_df')
		test = a.t_df.path.tolist()
		test = [tuple(path) for path in test]
		control = [(0,2,1),(2,1),(0,2)]
		check_pairs(control, test)

	# test update_loc_df_ids
	def test_update_loc_df_ids(self):
		a = get_dummy_sg()
		id_map = a.get_ordered_id_map()
		a.dfs_to_dicts()
		a.update_loc_df_ids(id_map)
		a.dicts_to_dfs()

		test = a.loc_df.apply(
			lambda x: (x.vertex_id, x.coord), axis=1)
		control = [(0,1),(1,2),(2,3)]
		check_pairs(control, test)

	# tests update_edge_df_ids
	def test_update_edge_df_ids(self):
		a = get_dummy_sg()
		id_map = a.get_ordered_id_map()
		a.dfs_to_dicts()
		a.update_edge_df_ids(id_map)
		a.dicts_to_dfs()

		test = a.edge_df.edge_id.tolist()
		control = [(0,1),(0,2),(2,1)]
		check_pairs(control, test)

	# tests update_t_df_paths
	def test_update_t_df_paths(self):
		a = get_dummy_sg()
		id_map = a.get_ordered_id_map()
		a.dfs_to_dicts()
		a.update_t_df_paths(id_map)
		a.dicts_to_dfs()

		test = a.t_df.path.tolist()
		test = [tuple(path) for path in test]
		control = [(0,2,1),(2,1),(0,2)]
		check_pairs(control, test)		

	# TODO extend this to work on multiple genes,
	# will require a more robust test
	# tests get_ordered_id_map
	def test_get_ordered_id_map(self):
		a = get_dummy_sg()
		id_map = a.get_ordered_id_map()

		test = list(id_map.items())
		control = [(0,0),(1,2),(2,1)]
		check_pairs(control, test)

	# tests create_graph_from_dfs
	def test_create_graph_from_dfs(self):
		### TODO ###
		pass
		### TODO ###

	# tests oder_edge_df
	def test_order_edge_df(self):
		a = get_dummy_sg()
		a.order_edge_df()

		v1 = a.edge_df.v1.tolist()
		assert v1 == sorted(v1)

	##########################################################################
	############################# Other utilities ############################
	##########################################################################

	# tests is_empty
	def test_is_empty(self):
		a = swan.SwanGraph()

		# 1: graph is empty, no datasets (including annotation) 
		print(a.datasets)
		assert a.is_empty() == True

		a.datasets = ['a']

		# 2: graph has one dataset
		print(a.datasets)
		assert a.is_empty() == False

		a.datasets = ['a', 'b']

		# 3: graph has more than one dataset
		print(a.datasets)
		assert a.is_empty() == False

	# tests get_dataset_cols
	def test_get_dataset_cols(self):
		a = swan.SwanGraph()

		# 1: graph is empty, no datasets (including annotation) 
		print(a.datasets)
		assert a.get_dataset_cols() == []

		a.datasets = ['a']

		# 2: graph has one dataset
		print(a.datasets)
		assert a.get_dataset_cols() == ['a']

		a.datasets = ['a', 'b']

		# 3: graph has more than one dataset
		print(a.datasets)
		assert a.get_dataset_cols() == ['a', 'b']

	# tests get_count_cols
	def test_get_count_cols(self):
		a = swan.SwanGraph()

		# 1: empty graph, no datasets argument
		assert a.get_count_cols() == []

		# 2: one dataset in graph, no datasets argument
		a.counts = ['a_counts']
		assert a.get_count_cols() == ['a_counts']

		# 3: > one dataset in graph, no datasets argument
		a.counts = ['a_counts', 'b_counts']
		assert a.get_count_cols() == ['a_counts', 'b_counts']

		# 4: empty graph, datasets argument
		# also tests ability to handle char input
		a = swan.SwanGraph()
		with pytest.raises(Exception) as excinfo:
			a.get_count_cols('a')
		assert 'Abundance for dataset a' in str(excinfo.value)

		# 5: graph with one dataset and datasets argument already in graph
		a.counts = ['a_counts']
		assert a.get_count_cols('a') == ['a_counts']

		# 6: graph with one dataset and datasets argument not in graph
		with pytest.raises(Exception) as excinfo:
			a.get_count_cols('b')
		assert 'Abundance for dataset b' in str(excinfo.value)

		# 7: graph with more than one dataset and datasets argument in graph
		a.counts = ['a_counts', 'b_counts']
		assert a.get_count_cols(['a', 'b']) == ['a_counts', 'b_counts']

		# 8: graph with more than one dataset and not all datasets argument in graph
		with pytest.raises(Exception) as excinfo:
			a.get_count_cols(['a', 'b', 'c'])
		assert 'Abundance for dataset c'

	# tests get_tpm_cols
	def test_get_tpm_cols(self):
		a = swan.SwanGraph()

		# 1: empty graph, no datasets argument
		assert a.get_tpm_cols() == []

		# 2: one dataset in graph, no datasets argument
		a.tpm = ['a_tpm']
		assert a.get_tpm_cols() == ['a_tpm']

		# 3: > one dataset in graph, no datasets argument
		a.tpm = ['a_tpm', 'b_tpm']
		assert a.get_tpm_cols() == ['a_tpm', 'b_tpm']

		# 4: empty graph, datasets argument
		# also tests ability to handle char input
		a = swan.SwanGraph()
		with pytest.raises(Exception) as excinfo:
			a.get_tpm_cols('a')
		assert 'Abundance for dataset a' in str(excinfo.value)

		# 5: graph with one dataset and datasets argument already in graph
		a.tpm = ['a_tpm']
		a.counts = ['a_counts']
		assert a.get_tpm_cols('a') == ['a_tpm']

		# 6: graph with one dataset and datasets argument not in graph
		with pytest.raises(Exception) as excinfo:
			a.get_tpm_cols('b')
		assert 'Abundance for dataset b' in str(excinfo.value)

		# 7: graph with more than one dataset and datasets argument in graph
		a.tpm = ['a_tpm', 'b_tpm']
		a.counts = ['a_counts', 'b_counts']
		assert a.get_tpm_cols(['a', 'b']) == ['a_tpm', 'b_tpm']

		# 8: graph with more than one dataset and not all datasets argument in graph
		with pytest.raises(Exception) as excinfo:
			a.get_tpm_cols(['a', 'b', 'c'])
		assert 'Abundance for dataset c'

	# tests get_strand_from_tid
	def test_get_strand_from_tid(self):
		a = get_dummy_sg()
		a.loc_df.loc[1, 'strand'] = '-'

		print(a.loc_df.head())

		assert a.get_strand_from_tid(0) == '+'
		assert a.get_strand_from_tid(1) == '-'
		assert a.get_strand_from_tid(2) == '+'

	# tests get_path_from_tid
	def test_get_path_from_tid(self):
		a = get_dummy_sg()

		assert a.get_path_from_tid(0) == [0,1]
		assert a.get_path_from_tid(1) == [1,2]
		assert a.get_path_from_tid(2) == [0,1,2]

	# tests get_gid_from_tid
	def test_get_gid_from_tid(self):
		a = get_dummy_sg()

		a.t_df.loc[1, 'gid'] = 3

		assert a.get_gid_from_tid(0) == 0
		assert a.get_gid_from_tid(1) == 3
		assert a.get_gid_from_tid(2) == 0

	# tests get_gene_min_max
	def test_get_gene_min_max(self):
		a = get_dummy_sg()
		assert a.get_gene_min_max(0) == (1, 3)

	# tests get_transcript_min_max
	def test_get_transcript_min_max(self):
		a = get_dummy_sg()
		assert a.get_transcript_min_max(0) == (1,3)
		assert a.get_transcript_min_max(1) == (2,3)
		assert a.get_transcript_min_max(2) == (1,2)

	# tests subset_on_gene
	def test_subset_on_gene(self):
		gid = 0
		a = swan.SwanGraph()
		a.t_df = pd.DataFrame({'tid': [0,1,2,3,4,5],
							   'gid': [0,0,1,1,2,2],
							   'path': [[0,1,2],
							   			[2,3,4],
							   			[5,6,7],
							   			[6,7,8],
							   			[9,10,11,12],
							   			[9,11,12]]})
		a.loc_df = pd.DataFrame({'vertex_id': [0,1,2,3,4,5,6,7,8,9,10,11,12], 
								 'strand': ['+','+','+','+','+','-','-','-','-','-','-','-','-'],
								 'chrom': [1,1,1,1,1,2,2,2,2,2,2,2,2],
								 'coord': [0,1,2,3,4,5,6,7,8,9,10,11,12]})
		a.edge_df = pd.DataFrame({'edge_id': [(0,1),(1,2),(2,3),(3,4),(5,6),(6,7),
											  (7,8),(9,10),(10,11),(11,12)]})
		a.t_df = swan.create_dupe_index(a.t_df, 'tid')
		a.t_df = swan.set_dupe_index(a.t_df, 'tid')
		a.loc_df = swan.create_dupe_index(a.loc_df, 'vertex_id')
		a.loc_df = swan.set_dupe_index(a.loc_df, 'vertex_id')
		a.edge_df = swan.create_dupe_index(a.edge_df, 'edge_id')
		a.edge_df = swan.set_dupe_index(a.edge_df, 'edge_id')

		# check subsetting for gene 0 
		a = swan.subset_on_gene(a, 0)

		test = a.t_df['tid'].tolist() 
		control = [0,1]
		check_pairs(control, test)

		test = a.loc_df['vertex_id'].tolist()
		control = [0,1,2,3,4]
		check_pairs(control, test)

		test = a.edge_df['edge_id'].tolist()
		control = [(0,1),(1,2),(2,3),(3,4)]
		check_pairs(control, test)

##########################################################################
############################# Pytest utilities ###########################
##########################################################################

def check_pairs(control, test):
	print('control')
	print(control)
	for t in test:
		print(t)
		assert t in control
	assert len(test) == len(control)

def get_dummy_sg(special=None):
	a = swan.SwanGraph()

	loc_df = pd.DataFrame({'chrom': [1,1,1],
		'coord': [1,3,2],
		'strand': ['+', '+', '+'],
		'vertex_id': [0,1,2]})
	loc_df = swan.create_dupe_index(loc_df, 'vertex_id')
	loc_df = swan.set_dupe_index(loc_df, 'vertex_id')

	edge_df = pd.DataFrame({'edge_id': [(0,2),(0,1),(1,2)],
			  'v1': [0,0,1],
			  'v2': [2,1,2],
			  'edge_type': ['exon', 'exon', 'intron'], 
			  'strand': ['+','+','+']})
	edge_df = swan.create_dupe_index(edge_df, 'edge_id')
	edge_df = swan.set_dupe_index(edge_df, 'edge_id')

	t_df = pd.DataFrame({'tid':[2,1,0],
			  'gid':[0,0,0],
			  'gname':['0','0','0'],
			  'path':[[0,1,2],[1,2],[0,1]],
			  'counts_a': [0,0,12],
			  'counts_b': [1,0,14]})
	t_df = swan.create_dupe_index(t_df, 'tid')
	t_df = swan.set_dupe_index(t_df, 'tid')

	if special == 'intron':
		edge_df.loc[(0,1), 'edge_type'] = 'intron'


	a.loc_df = loc_df
	a.edge_df = edge_df
	a.t_df = t_df 

	if special == 'no_locs':	
		pass
	else:
		a.get_loc_types()

	if special == 'dataset':
		a.datasets = ['dataset_a']
		a.loc_df['dataset_a'] = True
		a.edge_df['dataset_a'] = True
		a.t_df['dataset_a'] = True

	return a






